import datetime
from typing import Dict, ItemsView, List, Optional, Sequence, Set

from webviz_subsurface._providers import (
    EnsembleSummaryProvider,
    Frequency,
    VectorMetadata,
)


class EnsembleSummaryProviderSet:
    """
    Class to create a set of ensemble summary providers with unique names

    Provides interface for read-only fetching of provider data

    NOTE:
        - Assuming same implementation of EnsembleSummaryProvider interface across
        all providers in set. There is no guarantee of functionality if various
        provider implementations are mixed in the same set.
    """

    def __init__(self, provider_dict: Dict[str, EnsembleSummaryProvider]) -> None:
        self._provider_dict = provider_dict.copy()
        self._names = list(self._provider_dict.keys())
        self._all_vector_names = self._create_union_of_vector_names_from_providers(
            list(self._provider_dict.values())
        )
        self._all_realizations = self._create_union_of_realizations_from_providers(
            list(self._provider_dict.values())
        )

    def verify_consistent_vector_metadata(self) -> None:
        """
        Verify that vector metadata is consistent across providers, raise exception
        if inconsistency occur.

        TODO:
        * Improve print of inconsistent metadata info - store all inconsistencies
        and print, do not raise ValueError on first.
        * Replace with vector metadata dataclass object when updated (__eq__ operator
        for dataclass would be handy)
        """

        # Iterate through all vector names for provider set
        for vector_name in self._all_vector_names:
            # Store provider name and retrieved vector metadata for specific vector name
            vector_provider_metadata_dict: Dict[str, Optional[VectorMetadata]] = {}

            # Retrieve vector metadata from providers
            for name, provider in self._provider_dict.items():
                if vector_name in provider.vector_names():
                    vector_provider_metadata_dict[name] = provider.vector_metadata(
                        vector_name
                    )
            if vector_provider_metadata_dict:
                validator_provider, metadata_validator = list(
                    vector_provider_metadata_dict.items()
                )[0]
                for (
                    provider_name,
                    vector_metadata,
                ) in vector_provider_metadata_dict.items():
                    if vector_metadata != metadata_validator:

                        raise ValueError(
                            f'Inconsistent vector metadata for vector "{vector_name}"'
                            f' between provider "{validator_provider}" and provider '
                            f'{provider_name}"'
                        )

    @staticmethod
    def _create_union_of_vector_names_from_providers(
        providers: List[EnsembleSummaryProvider],
    ) -> List[str]:
        """Create list with the union of vector names among providers"""
        vector_names = []
        for provider in providers:
            vector_names.extend(provider.vector_names())
        vector_names = list(sorted(set(vector_names)))
        return vector_names

    @staticmethod
    def _create_union_of_realizations_from_providers(
        providers: Sequence[EnsembleSummaryProvider],
    ) -> List[int]:
        """Create list with the union of realizations among providers"""
        realizations: Set[int] = set()
        for provider in providers:
            realizations.update(provider.realizations())
        output = list(sorted(realizations))
        return output

    def items(self) -> ItemsView[str, EnsembleSummaryProvider]:
        return self._provider_dict.items()

    def provider_names(self) -> List[str]:
        return self._names

    def provider(self, name: str) -> EnsembleSummaryProvider:
        if name not in self._provider_dict.keys():
            raise ValueError(f'Provider with name "{name}" not present in set!')
        return self._provider_dict[name]

    def all_providers(self) -> List[EnsembleSummaryProvider]:
        return list(self._provider_dict.values())

    def all_dates(
        self,
        resampling_frequency: Optional[Frequency],
    ) -> List[datetime.datetime]:
        """List with the union of dates among providers"""
        dates_union: Set[datetime.datetime] = set()
        for provider in self.all_providers():
            _dates = set(provider.dates(resampling_frequency, None))
            dates_union.update(_dates)
        return list(sorted(dates_union))

    def all_realizations(self) -> List[int]:
        """List with the union of realizations among providers"""
        return self._all_realizations

    def all_vector_names(self) -> List[str]:
        """List with the union of vector names among providers"""
        return self._all_vector_names

    def vector_metadata(self, vector: str) -> Optional[VectorMetadata]:
        """Get vector metadata from first occurrence among providers,

        `return:`
        Vector metadata from first occurrence among providers, None if not existing
        """
        metadata: Optional[VectorMetadata] = next(
            (
                provider.vector_metadata(vector)
                for provider in self._provider_dict.values()
                if vector in provider.vector_names()
                and provider.vector_metadata(vector)
            ),
            None,
        )
        return metadata



import datetime
from typing import Optional

import numpy as np
import pandas as pd


def assert_date_column_is_datetime_object(df: pd.DataFrame) -> None:
    """Check if date column is on datetime.datetime format

    Raise ValueError if date column is missing, is empty or
    not on datetime.datetime format

    `Assume:`
    * Column named "DATE" exist
    * Each row element of "DATE" column is of same type. Mehtod utilize first
    element of "DATE" column as sample value to detect data type.

    `NOTE:`
    - When no rows - dtype of column is not checked, just early return!
    - Type check is performed using type() and not isinstance(), as
    isinstance() gives true on subclass. Thereby instance of pd.Timestamp
    return True - i.e. x: pd.Timestamp -> isinstance(x, datetime.datetime)
    is True.
    """
    if "DATE" not in df.columns:
        raise ValueError('df does not contain column "DATE"')

    # Empty rows (no dtype check of "DATE"-column)
    if df.shape[0] <= 0:
        return None

    # Get type from first element - use iloc to access by position rather than index label
    sampled_date_value = df["DATE"].iloc[0]

    # Use type() and not isinstance() as isinstance() gives true on subclass
    # pylint: disable = unidiomatic-typecheck
    if type(sampled_date_value) != datetime.datetime:
        raise ValueError(
            '"DATE"-column in dataframe is not on datetime.datetime format!'
        )

    return None


def make_date_column_datetime_object(df: pd.DataFrame) -> None:
    """Convert date column to datetime.datetime format

    Methods only handles "DATE" column of datetime.datetime or pd.Timestamp
    format. With date format datetime.datetime nothing is performed. If
    date column is of format pd.Timestamp the dates are converted. Otherwise
    ValueError is raised.

    `Assume:`
    * Column named "DATE" exist
    * "DATE" column is of type datetime.datetime or pd.Timestamp
    * Each row element of "DATE" column is of same type. Mehtod utilize first
    element of "DATE" column as sample value to detect data type.

    `NOTE:`
    - When no rows - dtype of column is not checked and converted, just early return!
    - Type check is performed using type() and not isinstance(), as
    isinstance() gives true on subclass. Thereby instance of pd.Timestamp
    return True - i.e. x: pd.Timestamp -> isinstance(x, datetime.datetime) is True
    """
    if "DATE" not in df.columns:
        raise ValueError('df does not contain column "DATE"')

    # Empty rows (no dtype check/conversion of "DATE"-column)
    if df.shape[0] <= 0:
        return None

    # Get type from first element - use iloc to access by position rather than index label
    sampled_date_value = df["DATE"].iloc[0]

    # Use type() and not isinstance() as isinstance() gives true on subclass
    # pylint: disable = unidiomatic-typecheck
    if type(sampled_date_value) == datetime.datetime:
        return None

    # Use type() and not isinstance() as isinstance() gives true on subclass
    # pylint: disable = unidiomatic-typecheck
    if type(sampled_date_value) == pd.Timestamp:
        df["DATE"] = pd.Series(
            df["DATE"].dt.to_pydatetime(), dtype=object, index=df.index
        )
        return None

    raise ValueError(
        f'Column "DATE" of type {type(sampled_date_value)} is not handled!'
    )


def correlate_response_with_dataframe(
    df: pd.DataFrame, response: str, corrwith: Optional[list] = None
) -> pd.Series:
    """Returns the correlation matrix for a dataframe soprted by correlation"""
    df = df[corrwith + [response]].copy() if corrwith is not None else df.copy()
    df = df[df.columns[df.nunique() > 1]]
    if response not in df.columns:
        df[response] = np.nan
    series = df[response]
    df = df.drop(columns=[response])
    corrdf = df.corrwith(series)
    return corrdf.reindex(corrdf.abs().sort_values().index)


def merge_dataframes_on_realization(
    dframe1: pd.DataFrame, dframe2: pd.DataFrame
) -> pd.DataFrame:
    """Merge two dataframes on realization"""
    dframe1.set_index("REAL", inplace=True)
    dframe2.set_index("REAL", inplace=True)
    return dframe1.join(dframe2).reset_index()



from typing import Dict, List, Optional

from webviz_config import WebvizSettings

from webviz_subsurface._providers import (
    EnsembleSurfaceProvider,
    EnsembleSurfaceProviderFactory,
    EnsembleTableProvider,
    EnsembleTableProviderFactory,
)
from webviz_subsurface._utils.webvizstore_functions import read_csv
from webviz_subsurface.plugins._co2_leakage._utilities.generic import MapAttribute
from webviz_subsurface.plugins._map_viewer_fmu._tmp_well_pick_provider import (
    WellPickProvider,
)


def init_map_attribute_names(
    mapping: Optional[Dict[str, str]]
) -> Dict[MapAttribute, str]:
    if mapping is None:
        # Based on name convention of xtgeoapp_grd3dmaps:
        return {
            MapAttribute.MIGRATION_TIME: "MigrationTime",
            MapAttribute.MAX_SGAS: "max_SGAS",
            MapAttribute.MAX_AMFG: "max_AMFG",
        }
    return {MapAttribute(key): value for key, value in mapping.items()}


def init_surface_providers(
    webviz_settings: WebvizSettings,
    ensembles: List[str],
) -> Dict[str, EnsembleSurfaceProvider]:
    surface_provider_factory = EnsembleSurfaceProviderFactory.instance()
    return {
        ens: surface_provider_factory.create_from_ensemble_surface_files(
            webviz_settings.shared_settings["scratch_ensembles"][ens],
        )
        for ens in ensembles
    }


def init_well_pick_provider(
    well_pick_path: Optional[str],
    map_surface_names_to_well_pick_names: Optional[Dict[str, str]],
) -> Optional[WellPickProvider]:
    if well_pick_path is None:
        return None
    try:
        return WellPickProvider(
            read_csv(well_pick_path), map_surface_names_to_well_pick_names
        )
    except OSError:
        return None


def init_co2_containment_table_providers(
    ensemble_roots: Dict[str, str],
    table_rel_path: str,
) -> Dict[str, EnsembleTableProvider]:
    return {
        ens: (
            EnsembleTableProviderFactory.instance().create_from_per_realization_csv_file(
                ens_path, table_rel_path
            )
        )
        for ens, ens_path in ensemble_roots.items()
    }



import logging
import shutil
from enum import Enum
from pathlib import Path
from typing import List, Optional

import numpy as np
import pandas as pd
import xtgeo

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._egrid_file_discovery import EclipseCaseFileInfo
from .ensemble_grid_provider import EnsembleGridProvider

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class Col:
    REAL = "realization"
    EGRID = "egrid_path"
    INIT = "init_path"
    UNRST = "unrst_path"


class GridType(str, Enum):
    GEOMETRY = "geometry"
    STATIC_PROPERTY = "static_property"
    DYNAMIC_PROPERTY = "dynamic_property"


class ProviderImplEgrid(EnsembleGridProvider):
    def __init__(
        self,
        provider_id: str,
        provider_dir: Path,
        grid_inventory_df: pd.DataFrame,
        init_properties: List[str],
        restart_properties: List[str],
    ) -> None:
        self._provider_id = provider_id
        self._provider_dir = provider_dir
        self._inventory_df = grid_inventory_df
        self._init_properties = init_properties
        self._restart_properties = restart_properties
        first_unrst = self._inventory_df[Col.UNRST][0]
        self._restart_dates = [
            str(dateint)
            for dateint in xtgeo.GridProperties.scan_dates(
                str(provider_dir / first_unrst), datesonly=True
            )
        ]

    @staticmethod
    def write_backing_store(
        storage_dir: Path,
        storage_key: str,
        eclipse_case_paths: List[EclipseCaseFileInfo],
        avoid_copying_grid_data: bool,
    ) -> None:
        """If avoid_copying_grid_data if True, the specified grid data will NOT be copied
        into the backing store, but will be referenced from their source locations.
        Note that this is only useful when running in non-portable mode and will fail
        in portable mode.
        """

        timer = PerfTimer()

        do_copy_grid_data_into_store = not avoid_copying_grid_data

        # All data for this provider will be stored inside a sub-directory
        # given by the storage key
        provider_dir = storage_dir / storage_key
        LOGGER.info(f"Writing grid data backing store to: {provider_dir}")
        provider_dir.mkdir(parents=True, exist_ok=True)

        ecl_stored_cases = []
        for ecl_case in eclipse_case_paths:

            if do_copy_grid_data_into_store:

                egrid_rel_path = (
                    f"{ecl_case.realization}-{Path(ecl_case.egrid_path).name}"
                )
                init_rel_path = (
                    f"{ecl_case.realization}-{Path(ecl_case.init_path).name}"
                )
                unrst_rel_path = (
                    f"{ecl_case.realization}-{Path(ecl_case.unrst_path).name}"
                )
                ecl_stored_cases.append(
                    EclipseCaseFileInfo(
                        realization=ecl_case.realization,
                        egrid_path=egrid_rel_path,
                        init_path=init_rel_path,
                        unrst_path=unrst_rel_path,
                    )
                )

                shutil.copyfile(ecl_case.egrid_path, provider_dir / egrid_rel_path)
                shutil.copyfile(ecl_case.init_path, provider_dir / init_rel_path)
                shutil.copyfile(ecl_case.unrst_path, provider_dir / unrst_rel_path)

            else:
                ecl_stored_cases.append(ecl_case)

        timer.lap_s()

        grid_inventory_df = pd.DataFrame(ecl_stored_cases)

        parquet_file_name = provider_dir / "grid_inventory.parquet"

        grid_inventory_df.to_parquet(path=parquet_file_name)

    @staticmethod
    def from_backing_store(
        storage_dir: Path,
        storage_key: str,
        init_properties: List[str],
        restart_properties: List[str],
    ) -> Optional["ProviderImplEgrid"]:

        provider_dir = storage_dir / storage_key
        parquet_file_name = provider_dir / "grid_inventory.parquet"

        try:
            grid_inventory_df = pd.read_parquet(path=parquet_file_name)

            return ProviderImplEgrid(
                storage_key,
                provider_dir,
                grid_inventory_df,
                init_properties,
                restart_properties,
            )

        except FileNotFoundError:
            return None

    def provider_id(self) -> str:
        return self._provider_id

    def static_property_names(self) -> List[str]:
        return self._init_properties

    def dynamic_property_names(self) -> List[str]:
        return self._restart_properties

    def dates_for_dynamic_property(self, property_name: str) -> Optional[List[str]]:
        return self._restart_dates

    def realizations(self) -> List[int]:
        unique_reals = self._inventory_df[Col.REAL].unique()

        # Sort and strip out any entries with real == -1
        return sorted([r for r in unique_reals if r >= 0])

    def get_3dgrid(self, realization: int) -> xtgeo.Grid:

        df = self._inventory_df.loc[self._inventory_df[Col.REAL] == realization]
        df = df[Col.EGRID]
        grid = xtgeo.grid_from_file(self._provider_dir / df.iloc[0], fformat="egrid")

        return grid

    def get_static_property_values(
        self, property_name: str, realization: int
    ) -> Optional[np.ndarray]:
        grid = self.get_3dgrid(realization)
        df = self._inventory_df.loc[self._inventory_df[Col.REAL] == realization]
        df = df[Col.INIT]
        grid_property = xtgeo.gridproperty_from_file(
            self._provider_dir / df.iloc[0],
            fformat="init",
            name=property_name,
            grid=grid,
        )
        fill_value = np.nan if not grid_property.isdiscrete else -1

        return grid_property.get_npvalues1d(order="F", fill_value=fill_value).ravel()

    def get_dynamic_property_values(
        self, property_name: str, property_date: str, realization: int
    ) -> Optional[np.ndarray]:
        grid = self.get_3dgrid(realization)
        df = self._inventory_df.loc[self._inventory_df[Col.REAL] == realization]
        df = df[Col.UNRST]
        grid_property = xtgeo.gridproperty_from_file(
            self._provider_dir / df.iloc[0],
            fformat="unrst",
            name=property_name,
            date=property_date,
            grid=grid,
        )
        return grid_property.get_npvalues1d(order="F").ravel()



########################################
#
#  Copyright (C) 2020-     Equinor ASA
#
########################################


from typing import Any, Callable, Dict, List, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from ..._datainput.pvt_data import load_pvt_csv, load_pvt_dataframe
from ._views._pvt import PvtView


class PvtPlot(WebvizPluginABC):
    """Visualizes formation volume factor and viscosity data \
    for oil, gas and water from both **csv**, Eclipse **init** and **include** files.

    !> The plugin supports variations in PVT between ensembles, but not between \
    realizations in the same ensemble.
    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`pvt_relative_file_path`:** Local path to a csv file in each \
        realization with dumped pvt data.
    * **`read_from_init_file`:** A boolean flag stating if data shall be \
        read from an Eclipse INIT file instead of an INCLUDE file. \
        This is only used when **pvt_relative_file_path** is not given.
    * **`drop_ensemble_duplicates`:** A boolean flag stating if ensembles \
        which are holding duplicate data of other ensembles shall be dropped. \
        Defaults to False.

    ---
    The minimum requirement is to define `ensembles`.

    If no `pvt_relative_file_path` is given, the PVT data will be extracted automatically
    from the simulation decks of individual realizations using `fmu_ensemble` and `ecl2df`.
    If the `read_from_init_file` flag is set to True, the extraction procedure in
    `ecl2df` will be replaced by an individual extracting procedure that reads the
    normalized Eclipse INIT file.
    Note that the latter two extraction methods can be very slow for larger data and are therefore
    not recommended unless you have a very simple model/data deck.
    If the `drop_ensemble_duplicates` flag is set to True, any ensembles which are holding
    duplicate data of other ensembles will be dropped.

    `pvt_relative_file_path` is a path to a file stored per realization (e.g. in \
    `share/results/tables/pvt.csv`). `pvt_relative_file_path` columns:
    * One column named `KEYWORD` or `TYPE`: with Flow/Eclipse style keywords
        (e.g. `PVTO` and `PVDG`).
    * One column named `PVTNUM` with integer `PVTNUM` regions.
    * One column named `RATIO` or `R` with the gas-oil-ratio as the primary variate.
    * One column named `PRESSURE` with the fluids pressure as the secondary variate.
    * One column named `VOLUMEFACTOR` as the first covariate.
    * One column named `VISCOSITY` as the second covariate.

    The file can e.g. be dumped to disc per realization by a forward model in ERT using
    `ecl2df`.
    """

    class Ids(StrEnum):
        INDICATORS = "pvt-indicators"

    PHASES = ["OIL", "GAS", "WATER"]

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: List[str],
        pvt_relative_file_path: str = None,
        read_from_init_file: bool = False,
        drop_ensemble_duplicates: bool = False,
    ) -> None:
        super().__init__(stretch=True)

        self.ensemble_paths = {
            ensemble: webviz_settings.shared_settings["scratch_ensembles"][ensemble]
            for ensemble in ensembles
        }

        self.plotly_theme = webviz_settings.theme.plotly_theme

        self.pvt_relative_file_path = pvt_relative_file_path

        self.read_from_init_file = read_from_init_file

        self.drop_ensemble_duplicates = drop_ensemble_duplicates

        if self.pvt_relative_file_path is None:
            self.pvt_df = load_pvt_dataframe(
                self.ensemble_paths,
                use_init_file=read_from_init_file,
                drop_ensemble_duplicates=drop_ensemble_duplicates,
            )
        else:
            # Load data from all ensembles into a pandas DataFrame
            self.pvt_df = load_pvt_csv(
                ensemble_paths=self.ensemble_paths,
                csv_file=self.pvt_relative_file_path,
                drop_ensemble_duplicates=drop_ensemble_duplicates,
            )

            self.pvt_df = self.pvt_df.rename(str.upper, axis="columns").rename(
                columns={"TYPE": "KEYWORD", "RS": "RATIO", "R": "RATIO", "GOR": "RATIO"}
            )

        # Ensure that the identifier string "KEYWORD" is contained in the header columns
        if "KEYWORD" not in self.pvt_df.columns:
            raise ValueError(
                (
                    "There has to be a KEYWORD or TYPE column with corresponding Eclipse keyword."
                    "When not providing a csv file, make sure ecl2df is installed."
                )
            )

        self.phases_additional_info: List[str] = []
        if self.pvt_df["KEYWORD"].str.contains("PVTO").any():
            self.phases_additional_info.append("PVTO")
        elif self.pvt_df["KEYWORD"].str.contains("PVDO").any():
            self.phases_additional_info.append("PVDO")
        elif self.pvt_df["KEYWORD"].str.contains("PVCDO").any():
            self.phases_additional_info.append("PVCDO")
        if self.pvt_df["KEYWORD"].str.contains("PVTG").any():
            self.phases_additional_info.append("PVTG")
        elif self.pvt_df["KEYWORD"].str.contains("PVDG").any():
            self.phases_additional_info.append("PVDG")
        if self.pvt_df["KEYWORD"].str.contains("PVTW").any():
            self.phases_additional_info.append("PVTW")

        self.add_view(
            PvtView(self.pvt_df, webviz_settings),
            PvtPlot.Ids.INDICATORS,
        )

    def add_webvizstore(
        self,
    ) -> List[Tuple[Callable, List[Dict[str, Any]]]]:
        return (
            [
                (
                    load_pvt_dataframe,
                    [
                        {
                            "ensemble_paths": self.ensemble_paths,
                            "use_init_file": self.read_from_init_file,
                            "drop_ensemble_duplicates": self.drop_ensemble_duplicates,
                        }
                    ],
                )
            ]
            if self.pvt_relative_file_path is None
            else [
                (
                    load_pvt_csv,
                    [
                        {
                            "ensemble_paths": self.ensemble_paths,
                            "csv_file": self.pvt_relative_file_path,
                            "drop_ensemble_duplicates": self.drop_ensemble_duplicates,
                        }
                    ],
                )
            ]
        )



from typing import Dict, List

import numpy as np
import pandas as pd
from webviz_config.common_cache import CACHE

from ...._utils.fanchart_plotting import (
    FanchartData,
    FreeLineData,
    LowHighData,
    MinMaxData,
    get_fanchart_traces,
)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_df(df: pd.DataFrame, ensemble: str, wells: List[str]) -> pd.DataFrame:
    """Filter dataframe for current ensembles and wells.
    Replacing zeroes (well not open) with np.NaN to not be accounted for
    in statistics.
    """
    columns = ["ENSEMBLE"] + [f"WBHP:{well}" for well in wells]
    return df.loc[df["ENSEMBLE"] == ensemble][columns].replace(0, np.NaN)


def calc_statistics(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate statistics for given vectors over the ensembles
    refaxis is used if another column than DATE should be used to groupby.
    """
    # Invert p10 and p90 due to oil industry convention.
    def p10(x: np.ndarray) -> float:
        return np.nanpercentile(x, q=90)

    def p50(x: np.ndarray) -> float:
        return np.nanpercentile(x, q=50)

    def p90(x: np.ndarray) -> float:
        return np.nanpercentile(x, q=10)

    # Calculate statistics, ignoring NaNs.
    stat_df = (
        df.groupby("ENSEMBLE")
        .agg([np.nanmean, "count", np.nanstd, np.nanmin, np.nanmax, p10, p50, p90])
        .reset_index(drop=True, level="ENSEMBLE")
    )
    # Rename nanmin, nanmax and nanmean to min, max and mean.
    col_stat_label_map = {
        "nanmin": "min",
        "nanmax": "max",
        "nanmean": "mean",
        "nanstd": "std",
        "p10": "high_p10",
        "p90": "low_p90",
    }
    stat_df.rename(columns=col_stat_label_map, level=1, inplace=True)
    stat_df = stat_df.transpose().unstack()
    stat_df.columns = stat_df.columns.get_level_values(1)  # Remove 0 index column
    return stat_df


def _get_fanchart_traces(
    ens_stat_df: pd.DataFrame,
    color: str,
    legend_group: str,
) -> List[dict]:
    """Renders a fanchart for an ensemble vector."""

    x = [vec[5:] for vec in ens_stat_df.index]

    data = FanchartData(
        samples=x,
        low_high=LowHighData(
            low_data=ens_stat_df["low_p90"].values,
            low_name="P90 (low)",
            high_data=ens_stat_df["high_p10"].values,
            high_name="P10 (high)",
        ),
        minimum_maximum=MinMaxData(
            minimum=ens_stat_df["min"].values,
            maximum=ens_stat_df["max"].values,
        ),
        free_line=FreeLineData("Mean", ens_stat_df["mean"].values),
    )

    return get_fanchart_traces(
        data=data,
        hex_color=color,
        legend_group=legend_group,
        hovermode="x",
    )


# Properties


def label_map() -> Dict[str, str]:
    """returns a map of labels for statistics."""
    return {
        "Mean": "mean",
        "Count (data points)": "count",
        "Stddev": "std",
        "Minimum": "min",
        "Maximum": "max",
        "P10 (high)": "high_p10",
        "P50": "p50",
        "P90 (low)": "low_p90",
    }



import datetime
import logging
import time
from pathlib import Path
from typing import Optional

from .ensemble_summary_provider import EnsembleSummaryProvider, Frequency
from .ensemble_summary_provider_factory import EnsembleSummaryProviderFactory


def _get_n_vectors_one_by_one_all_realizations(
    provider: EnsembleSummaryProvider,
    resampling_frequency: Optional[Frequency],
    num_vectors: int,
) -> None:

    all_vectors = provider.vector_names()
    num_vectors = min(num_vectors, len(all_vectors))
    vectors_to_get = all_vectors[0:num_vectors]

    print("## ------------------")
    print(
        f"## entering _get_n_vectors_one_by_one_all_realizations("
        f"{resampling_frequency}, {num_vectors}) ..."
    )

    start_tim = time.perf_counter()

    accum_rows = 0
    accum_cols = 0
    for vec_name in vectors_to_get:
        df = provider.get_vectors_df([vec_name], resampling_frequency)
        accum_rows += df.shape[0]
        accum_cols += df.shape[1]

    elapsed_time_ms = 1000 * (time.perf_counter() - start_tim)

    avg_shape = (
        int(accum_rows / num_vectors),
        int(accum_cols / num_vectors),
    )
    print("## avg shape:", avg_shape)
    print("## num unique dates:", df["DATE"].nunique())
    print("## num unique realizations:", df["REAL"].nunique())
    print("## avg time per vector (ms):", elapsed_time_ms / num_vectors)
    print(
        "## finished _get_n_vectors_one_by_one_all_realizations(), total time (ms)",
        elapsed_time_ms,
    )
    print("## ------------------")


def _get_n_vectors_in_batch_all_realizations(
    provider: EnsembleSummaryProvider,
    resampling_frequency: Optional[Frequency],
    num_vectors: int,
) -> None:

    all_vectors = provider.vector_names()
    num_vectors = min(num_vectors, len(all_vectors))
    vectors_to_get = all_vectors[0:num_vectors]

    print("## ------------------")
    print(
        f"## entering _get_n_vectors_in_batch_all_realizations("
        f"{resampling_frequency}, {num_vectors}) ..."
    )

    start_tim = time.perf_counter()

    df = provider.get_vectors_df(vectors_to_get, resampling_frequency)

    elapsed_time_ms = 1000 * (time.perf_counter() - start_tim)

    print("## df shape:", df.shape)
    print("## num unique dates:", df["DATE"].nunique())
    print("## num unique realizations:", df["REAL"].nunique())

    print("## avg time per vector (ms):", elapsed_time_ms / num_vectors)
    print(
        "## finished _get_n_vectors_in_batch_all_realizations(), total time (ms)",
        elapsed_time_ms,
    )
    print("## ------------------")


def _get_n_vectors_for_date_all_realizations(
    provider: EnsembleSummaryProvider, num_vectors: int, date: datetime.datetime
) -> None:

    all_vectors = provider.vector_names()
    num_vectors = min(num_vectors, len(all_vectors))
    vectors_to_get = all_vectors[0:num_vectors]

    print("## ------------------")
    print(
        f"## entering _get_n_vectors_for_date_all_realizations({num_vectors}, {date}) ..."
    )

    start_tim = time.perf_counter()

    df = provider.get_vectors_for_date_df(date, vectors_to_get)

    elapsed_time_ms = 1000 * (time.perf_counter() - start_tim)

    print("## df shape:", df.shape)
    print("## num unique realizations:", df["REAL"].nunique())
    print(
        "## finished _get_n_vectors_for_date_all_realizations(), total time (ms)",
        elapsed_time_ms,
    )
    print("## ------------------")


def _get_vector_names_filtered_by_value(provider: EnsembleSummaryProvider) -> None:

    print("## ------------------")
    print("## entering _get_vector_names_filtered_by_value() ...")

    start_tim = time.perf_counter()

    filtered_vec_names = provider.vector_names_filtered_by_value(
        exclude_all_values_zero=True, exclude_constant_values=True
    )
    print("## number of names returned:", len(filtered_vec_names))

    elapsed_time_ms = 1000 * (time.perf_counter() - start_tim)

    print(
        "## finished _get_vector_names_filtered_by_value(), total time (ms)",
        elapsed_time_ms,
    )
    print("## ------------------")


def _get_meta_for_n_vectors_one_by_one(
    provider: EnsembleSummaryProvider, num_vectors: int
) -> None:

    all_vectors = provider.vector_names()
    num_vectors = min(num_vectors, len(all_vectors))
    vectors_to_get = all_vectors[0:num_vectors]

    print("## ------------------")
    print(f"## entering _get_meta_for_n_vectors_one_by_one({num_vectors}) ...")

    start_tim = time.perf_counter()

    for vec_name in vectors_to_get:
        meta_dict = provider.vector_metadata(vec_name)
        print(f"metadata for {vec_name}")
        print(meta_dict)

    elapsed_time_ms = 1000 * (time.perf_counter() - start_tim)

    print(
        "## finished _get_meta_for_n_vectors_one_by_one(), total time (ms)",
        elapsed_time_ms,
    )
    print("## ------------------")


def _run_perf_tests(
    provider: EnsembleSummaryProvider, resampling_frequency: Optional[Frequency]
) -> None:

    all_vector_names = provider.vector_names()
    all_realizations = provider.realizations()
    all_dates = provider.dates(resampling_frequency)
    num_vectors = len(all_vector_names)
    num_realizations = len(all_realizations)
    num_dates = len(all_dates)
    print("## num_vectors:", num_vectors)
    print(f"## num_realizations: {num_realizations}   (head:{all_realizations[0:5]})")
    print(f"## num_dates: {num_dates}   (head: {all_dates[0:5]})")

    print()

    _get_meta_for_n_vectors_one_by_one(provider, 10)

    _get_vector_names_filtered_by_value(provider)

    _get_n_vectors_one_by_one_all_realizations(provider, resampling_frequency, 1)
    _get_n_vectors_one_by_one_all_realizations(provider, resampling_frequency, 50)

    _get_n_vectors_in_batch_all_realizations(provider, resampling_frequency, 50)
    _get_n_vectors_in_batch_all_realizations(provider, resampling_frequency, 99999)

    num_vecs = 100
    _get_n_vectors_for_date_all_realizations(
        provider, num_vecs, all_dates[int(num_dates / 2)]
    )
    _get_n_vectors_for_date_all_realizations(
        provider, num_vecs, all_dates[int(num_dates / 4)]
    )

    num_vecs = 99999
    _get_n_vectors_for_date_all_realizations(provider, num_vecs, all_dates[0])
    _get_n_vectors_for_date_all_realizations(provider, num_vecs, all_dates[-1])
    _get_n_vectors_for_date_all_realizations(
        provider, num_vecs, all_dates[int(num_dates / 2)]
    )
    _get_n_vectors_for_date_all_realizations(
        provider, num_vecs, all_dates[int(num_dates / 4)]
    )


def main() -> None:
    print()
    print("## Running EnsembleSummaryProvider performance tests")
    print("## =================================================")

    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s %(levelname)-3s [%(name)s]: %(message)s",
    )
    logging.getLogger("webviz_subsurface").setLevel(level=logging.INFO)
    logging.getLogger("webviz_subsurface").setLevel(level=logging.DEBUG)

    root_storage_dir = Path("/home/sigurdp/buf/webviz_storage_dir")
    # root_storage_dir = Path("/home/sigurdp/buf/blobfuse/mounted/webviz_storage_dir")

    # pylint: disable=line-too-long
    ensemble_path = "/home/sigurdp/gitRoot/webviz-subsurface-testdata/01_drogon_ahm/realization-*/iter-0"
    # ensemble_path = "/home/sigurdp/gitRoot/webviz-subsurface-testdata/01_drogon_design/realization-*/iter-0"
    # ensemble_path = "/home/sigurdp/webviz_testdata/reek_history_match_reduced/realization-*/iter-0"
    # ensemble_path = "/home/sigurdp/webviz_testdata/reek_history_match_large/realization-*/iter-0"

    frequency: Optional[Frequency] = Frequency.DAILY

    print()
    print("## root_storage_dir:", root_storage_dir)
    print("## ensemble_path:", ensemble_path)
    print("## frequency:", frequency)

    print()
    print("## Creating EnsembleSummaryProvider...")

    start_tim = time.perf_counter()
    factory = EnsembleSummaryProviderFactory(
        root_storage_dir, allow_storage_writes=True
    )

    provider = factory.create_from_arrow_unsmry_lazy(
        ens_path=ensemble_path, rel_file_pattern="share/results/unsmry/*.arrow"
    )
    resampling_frequency = frequency

    # provider = factory.create_from_arrow_unsmry_presampled(ensemble_path, frequency)
    # resampling_frequency = None

    print(
        "## Done creating EnsembleSummaryProvider, elapsed time (s):",
        time.perf_counter() - start_tim,
    )

    print()
    print("## Running perf tests...")
    _run_perf_tests(provider, resampling_frequency)

    print("## done")


# Running:
#   python -m webviz_subsurface._providers.ensemble_summary_provider.dev_provider_perf_testing
# -------------------------------------------------------------------------
if __name__ == "__main__":
    main()



from typing import Type

from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC
from webviz_core_components import Graph as WccGraph


class Graph(ViewElementABC):
    class Ids(StrEnum):
        GRAPH = "graph"

    def __init__(self, height: str = "90vh") -> None:
        super().__init__()

        self.height = height

    def inner_layout(self) -> Type[Component]:
        return WccGraph(
            id=self.register_component_unique_id(Graph.Ids.GRAPH),
            style={"height": self.height, "min-height": "300px"},
        )



from typing import TypedDict

from webviz_config.utils import StrEnum


class DeltaEnsemble(TypedDict):
    """Definition of delta ensemble

    Pair of names representing a delta ensemble: A-B
    """

    ensemble_a: str
    ensemble_b: str


class FanchartOptions(StrEnum):
    """
    Type definition for statistical options for fanchart
    """

    MEAN = "Mean"  # Mean value
    MIN_MAX = "Min/Max"  # Minimum and maximum pair
    P10_P90 = "P10/P90"  # P10 and P90 pair


class StatisticsFromOptions(StrEnum):
    """
    Type definition of options for what to generate statistics from in simulation time series
    """

    ALL_REALIZATIONS = "all_realizations"
    SELECTED_REALIZATIONS = "selected_realizations"


class StatisticsOptions(StrEnum):
    """
    Type definition of statistics to include in plot in simulation time series
    """

    MEAN = "Mean"
    MIN = "Min"
    MAX = "Max"
    P10 = "P10"
    P90 = "P90"
    P50 = "P50"


class SubplotGroupByOptions(StrEnum):
    """
    Type definition of options for subplots "group by" in graph for simulation time series
    """

    VECTOR = "vector"
    ENSEMBLE = "ensemble"


class TraceOptions(StrEnum):
    """
    Type definition for trace options in simulation time series
    """

    HISTORY = "history"
    OBSERVATIONS = "observations"


class VisualizationOptions(StrEnum):
    """
    Type definition for visualization options in simulation time series
    """

    REALIZATIONS = "realizations"
    STATISTICS = "statistics"
    FANCHART = "fanchart"
    STATISTICS_AND_REALIZATIONS = "statistics and realizations"



from typing import Dict, List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class ViewSettings(SettingsGroupABC):
    class Ids(StrEnum):
        SHOW_PLOTS = "show-plots"

    def __init__(self) -> None:
        super().__init__("Show Plots")

    @staticmethod
    def plot_visibility_options(phase: str = "") -> Dict[str, str]:
        options = {
            "fvf": "Formation Volume Factor",
            "viscosity": "Viscosity",
            "density": "Density",
            "ratio": "Gas/Oil Ratio (Rs)",
        }
        if phase == "OIL":
            options["ratio"] = "Gas/Oil Ratio (Rs)"
        if phase == "GAS":
            options["ratio"] = "Vaporized Oil Ratio (Rv)"
        if phase == "WATER":
            options.pop("ratio")
        return options

    def layout(self) -> List[Component]:
        component_id = self.register_component_unique_id(ViewSettings.Ids.SHOW_PLOTS)
        return [
            wcc.Checklist(
                id={"id": component_id, "plot": plot_value},
                options=[{"label": plot_label, "value": plot_value}],
                value=[plot_value],
                persistence=False,
            )
            for plot_value, plot_label in self.plot_visibility_options().items()
        ]



import glob
import io
import json
import logging
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

from webviz_config.webviz_store import webvizstore


class StratigraphyModel:
    """Facilitates loading of a json file with stratigraphy.

    The file needs to follow the format below. It is a tree structure, where each
    node has a name, an optional `color` parameter, and an optional `subzones`
    parameter which itself is a list of the same format.

    It is assumed that the file is the same for all realizations, so it will only
    read the first file it finds.


    ```json
    [
        {
            "name": "ZoneA",
            "color": "#FFFFFF",
            "subzones": [
                {
                    "name": "ZoneA.1"
                },
                {
                    "name": "ZoneA.2"
                }
            ]
        },
        {
            "name": "ZoneB",
            "color": "#FFF000",
            "subzones": [
                {
                    "name": "ZoneB.1",
                    "color": "#FFF111"
                },
                {
                    "name": "ZoneB.2",
                    "subzones: {"name": "ZoneB.2.2"}
                }
            ]
        },
    ]
    """

    def __init__(self, ens_name: str, ens_path: Path, stratigraphy_file: str):
        self._ens_name = ens_name
        self._ens_path = ens_path
        self._stratigraphy_file = stratigraphy_file
        self._data = list(json.load(self._load_data()))

    def __repr__(self) -> str:
        """This is necessary for webvizstore to work on objects"""
        return f"""
StratigraphyModel({self._ens_name!r}, {self._ens_path!r}, {self._stratigraphy_file!r})
        """

    @property
    def data(self) -> Optional[List[Dict[str, Any]]]:
        """Returns the stratigraphy if the list is not empty. If it is empty return None."""
        if self._data:
            return self._data
        return None

    @property
    def webviz_store(self) -> Tuple[Callable, List[Dict]]:
        return (
            self._load_data,
            [
                {
                    "self": self,
                }
            ],
        )

    @webvizstore
    def _load_data(self) -> io.BytesIO:
        """Reads the stratigraphy file for an ensemble. It returns the data from the first
        file it finds so it is assumed that the stratigraphy is equal for all realizations.

        If no file is found, it returns a empty list. Because of webvizstore it is not possible
        to return None.
        """

        for filename in glob.glob(f"{self._ens_path}/{self._stratigraphy_file}"):
            file_content = json.loads(Path(filename).read_text())
            logging.debug(f"Stratigraphy loaded from file: {filename}")
            return io.BytesIO(json.dumps(file_content).encode())
        return io.BytesIO(json.dumps("").encode())



from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import pandas as pd
from webviz_config.utils import StrEnum
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._datainput.fmu_input import scratch_ensemble


class TreeType(StrEnum):
    GRUPTREE = "GRUPTREE"
    BRANPROP = "BRANPROP"


class GruptreeModel:
    """Facilitates loading of gruptree tables. Can be reused in all
    plugins that are using grouptree data and extended with additional
    functionality and filtering options if necessary.
    """

    def __init__(
        self,
        ens_name: str,
        ens_path: Path,
        gruptree_file: str,
        tree_type: Optional[str] = None,
    ):
        self._ens_name = ens_name
        self._ens_path = ens_path
        self._gruptree_file = gruptree_file
        self._tree_type = TreeType(tree_type) if tree_type is not None else None
        self._dataframe = self.read_ensemble_gruptree()

        self._gruptrees_are_equal_over_reals = (
            self._dataframe["REAL"].nunique() == 1
            if not self._dataframe.empty
            else False
        )

    @property
    def dataframe(self) -> pd.DataFrame:
        """Returns a dataframe that will have the following columns:
        * DATE
        * CHILD (node in tree)
        * PARENT (node in tree)
        * KEYWORD (GRUPTREE, WELSPECS or BRANPROP)
        * REAL

        If gruptrees are exactly equal in all realizations then only one tree is
        stored in the dataframe. That means the REAL column will only have one unique value.
        If not, all trees are stored.
        """
        return self._dataframe

    def get_filtered_dataframe(
        self,
        terminal_node: Optional[str] = None,
        excl_well_startswith: Optional[List[str]] = None,
        excl_well_endswith: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """This function returns a sub-set of the rows in the gruptree dataframe
        filtered according to the input arguments:

        - terminal_node: returns the terminal node and all nodes below it in the
        tree (for all realizations and dates)
        - excl_well_startswith: removes WELSPECS rows where CHILD starts with any
        of the entries in the list.
        - excl_well_endswith: removes WELSPECS rows where CHILD ends with any
        of the entries in the list.

        """
        df = self._dataframe

        if terminal_node is not None:

            if terminal_node not in self._dataframe["CHILD"].unique():
                raise ValueError(
                    f"Terminal node '{terminal_node}' not found in 'CHILD' column "
                    "of the gruptree data."
                )
            if terminal_node != "FIELD":
                branch_nodes = self._get_branch_nodes(terminal_node)
                df = self._dataframe[self._dataframe["CHILD"].isin(branch_nodes)]

        def filter_wells(
            dframe: pd.DataFrame, well_name_criteria: Callable
        ) -> pd.DataFrame:
            return dframe[
                (dframe["KEYWORD"] != "WELSPECS")
                | (
                    (dframe["KEYWORD"] == "WELSPECS")
                    & (~well_name_criteria(dframe["CHILD"]))
                )
            ]

        if excl_well_startswith is not None:
            # Filter out WELSPECS rows where CHILD starts with any element in excl_well_startswith
            # Conversion to tuple done outside lambda due to mypy
            excl_well_startswith_tuple = tuple(excl_well_startswith)
            df = filter_wells(
                df, lambda x: x.str.startswith(excl_well_startswith_tuple)
            )

        if excl_well_endswith is not None:
            # Filter out WELSPECS rows where CHILD ends with any element in excl_well_endswith
            # Conversion to tuple done outside lambda due to mypy
            excl_well_endswith_tuple = tuple(excl_well_endswith)
            df = filter_wells(df, lambda x: x.str.endswith(excl_well_endswith_tuple))

        return df.copy()

    def _get_branch_nodes(self, terminal_node: str) -> List[str]:
        """The function is using recursion to find all wells below the node
        in the three.
        """
        branch_nodes = [terminal_node]

        children = self._dataframe[
            self._dataframe["PARENT"] == terminal_node
        ].drop_duplicates(subset=["CHILD"], keep="first")

        for _, childrow in children.iterrows():
            branch_nodes.extend(self._get_branch_nodes(childrow["CHILD"]))
        return branch_nodes

    @property
    def gruptrees_are_equal_over_reals(self) -> bool:
        """Returns true if gruptrees are exactly equal in all realizations."""
        return self._gruptrees_are_equal_over_reals

    def __repr__(self) -> str:
        """This is necessary for webvizstore to work on objects"""
        return f"""
GruptreeDataModel({self._ens_name!r}, {self._ens_path!r}, {self._gruptree_file!r})
        """

    @property
    def webviz_store(self) -> Tuple[Callable, List[Dict]]:
        return (
            self.read_ensemble_gruptree,
            [
                {
                    "self": self,
                }
            ],
        )

    @webvizstore
    def read_ensemble_gruptree(
        self, df_files: Optional[pd.DataFrame] = None
    ) -> pd.DataFrame:
        """Reads the gruptree files for an ensemble from the scratch disk. These
        files can be exported in the FMU workflow using the ECL2CSV
        forward model with subcommand gruptree.

        If tree_type == BRANPROP then GRUPTREE rows are filtered out
        If tree_type == GRUPTREE then BRANPROP rows are filtered out

        If the trees are equal in every realization, only one realization is kept.

        It is possible to pass a dataframe of file names (only columns required is
        REAL and FULLPATH). This is mostly intended for testing. If this is defaulted
        the files are found automatically using the scratch_ensemble.
        """
        if df_files is None:
            ens = scratch_ensemble(
                self._ens_name, str(self._ens_path), filter_file="OK"
            )
            df_files = ens.find_files(self._gruptree_file)

        if df_files.empty:
            return pd.DataFrame()

        # Load all gruptree dataframes and check if they are equal
        compare_columns = ["DATE", "CHILD", "KEYWORD", "PARENT"]
        df_prev = pd.DataFrame()
        dataframes = []
        gruptrees_are_equal = True
        for i, row in df_files.iterrows():
            df_real = pd.read_csv(row["FULLPATH"])
            unique_keywords = df_real["KEYWORD"].unique()

            if self._tree_type is None:
                # if tree_type is None, then we filter out GRUPTREE if BRANPROP
                # exists, if else we do nothing.
                if TreeType.BRANPROP.value in unique_keywords:
                    df_real = df_real[df_real["KEYWORD"] != TreeType.GRUPTREE.value]

            else:
                if self._tree_type.value not in unique_keywords:
                    raise ValueError(
                        f"Keyword {self._tree_type.value} not found in {row['FULLPATH']}"
                    )
                if (
                    self._tree_type == TreeType.GRUPTREE
                    and TreeType.BRANPROP.value in unique_keywords
                ):
                    # Filter out BRANPROP entries
                    df_real = df_real[df_real["KEYWORD"] != TreeType.BRANPROP.value]

                if self._tree_type == TreeType.BRANPROP:
                    # Filter out GRUPTREE entries
                    df_real = df_real[df_real["KEYWORD"] != TreeType.GRUPTREE.value]

            if (
                i > 0
                and gruptrees_are_equal
                and not df_real[compare_columns].equals(df_prev)
            ):
                gruptrees_are_equal = False
            else:
                df_prev = df_real[compare_columns].copy()

            df_real["REAL"] = row["REAL"]
            dataframes.append(df_real)
        df = pd.concat(dataframes)

        # Return either one or all realization in a common dataframe
        if gruptrees_are_equal:
            df = df[df["REAL"] == df["REAL"].min()]

        df["DATE"] = pd.to_datetime(df["DATE"])

        return df.where(pd.notnull(df), None)



from datetime import date, datetime
from pathlib import Path

import pandas as pd
import pytest
from _pytest.fixtures import SubRequest

from webviz_subsurface._providers.ensemble_summary_provider._provider_impl_arrow_presampled import (
    ProviderImplArrowPresampled,
)
from webviz_subsurface._providers.ensemble_summary_provider.ensemble_summary_provider import (
    EnsembleSummaryProvider,
)

# fmt: off
INPUT_DATA_DATETIME = [
    ["DATE",                          "REAL",  "A",  "C",   "Z"],
    [datetime(2021, 12, 20, 23, 59),  0,       10.0,  1.0,  0.0 ],
    [datetime(2021, 12, 20, 23, 59),  1,       12.0,  1.0,  0.0 ],
    [datetime(2021, 12, 21, 22, 58),  1,       13.0,  1.0,  0.0 ],
]

INPUT_DATA_AFTER_2262 = [
    ["DATE",                          "REAL",  "A",  "C",   "Z"],
    [datetime(2500, 12, 20, 23, 59),  0,       10.0,  1.0,  0.0 ],
    [datetime(2500, 12, 20, 23, 59),  1,       12.0,  1.0,  0.0 ],
    [datetime(2500, 12, 21, 22, 58),  1,       13.0,  1.0,  0.0 ],
]

INPUT_DATA_DATE = [
    ["DATE",              "REAL",  "A",  "C",   "Z"],
    [date(2022, 12, 20),  0,       10.0,  1.0,  0.0 ],
    [date(2022, 12, 20),  1,       12.0,  1.0,  0.0 ],
    [date(2022, 12, 21),  1,       13.0,  1.0,  0.0 ],
]

INPUT_DATA_STR = [
    ["DATE",       "REAL",  "A",  "C",   "Z"],
    ["2023-12-20",  0,      10.0,  1.0,  0.0 ],
    ["2023-12-20",  1,      12.0,  1.0,  0.0 ],
    ["2023-12-21",  1,      13.0,  1.0,  0.0 ],
]
# fmt: on


@pytest.fixture(
    name="provider",
    params=[
        INPUT_DATA_DATETIME,
        INPUT_DATA_AFTER_2262,
        INPUT_DATA_DATE,
        INPUT_DATA_STR,
    ],
)
def fixture_provider(request: SubRequest, tmp_path: Path) -> EnsembleSummaryProvider:

    input_py = request.param
    storage_dir = tmp_path

    input_df = pd.DataFrame(input_py[1:], columns=input_py[0])

    ProviderImplArrowPresampled.write_backing_store_from_ensemble_dataframe(
        storage_dir, "dummy_key", input_df
    )
    new_provider = ProviderImplArrowPresampled.from_backing_store(
        storage_dir, "dummy_key"
    )

    if not new_provider:
        raise ValueError("Failed to create EnsembleSummaryProvider")

    return new_provider


def test_get_vector_names(provider: EnsembleSummaryProvider) -> None:

    all_vecnames = provider.vector_names()
    assert len(all_vecnames) == 3
    assert all_vecnames == ["A", "C", "Z"]

    non_const_vec_names = provider.vector_names_filtered_by_value(
        exclude_constant_values=True
    )
    assert len(non_const_vec_names) == 1
    assert non_const_vec_names == ["A"]

    non_zero_vec_names = provider.vector_names_filtered_by_value(
        exclude_all_values_zero=True
    )
    assert len(non_zero_vec_names) == 2
    assert non_zero_vec_names == ["A", "C"]


def test_get_realizations(provider: EnsembleSummaryProvider) -> None:

    all_realizations = provider.realizations()
    assert len(all_realizations) == 2


def test_get_dates(provider: EnsembleSummaryProvider) -> None:

    intersection_of_dates = provider.dates(resampling_frequency=None)
    assert len(intersection_of_dates) == 1
    assert isinstance(intersection_of_dates[0], datetime)

    r0_dates = provider.dates(resampling_frequency=None, realizations=[0])
    r1_dates = provider.dates(resampling_frequency=None, realizations=[1])
    assert len(r0_dates) == 1
    assert len(r1_dates) == 2


def test_get_vectors(provider: EnsembleSummaryProvider) -> None:

    all_vecnames = provider.vector_names()
    assert len(all_vecnames) == 3

    vecdf = provider.get_vectors_df(["A"], resampling_frequency=None)
    assert vecdf.shape == (3, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "A"]

    sampleddate = vecdf["DATE"][0]
    assert isinstance(sampleddate, datetime)

    vecdf = provider.get_vectors_df(["A"], resampling_frequency=None, realizations=[1])
    assert vecdf.shape == (2, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "A"]

    vecdf = provider.get_vectors_df(
        ["C", "A"], resampling_frequency=None, realizations=[0]
    )
    assert vecdf.shape == (1, 4)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "C", "A"]


def test_get_vectors_for_date(provider: EnsembleSummaryProvider) -> None:

    intersection_of_dates = provider.dates(resampling_frequency=None)
    assert len(intersection_of_dates) == 1

    date_to_get = intersection_of_dates[0]
    assert isinstance(date_to_get, datetime)

    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A"])
    assert vecdf.shape == (2, 2)
    assert vecdf.columns.tolist() == ["REAL", "A"]

    date_to_get = intersection_of_dates[0]
    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A", "C"], [0])
    assert vecdf.shape == (1, 3)
    assert vecdf.columns.tolist() == ["REAL", "A", "C"]

    date_to_get = intersection_of_dates[0]
    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A", "Z"], [0])
    assert vecdf.shape == (1, 3)
    assert vecdf.columns.tolist() == ["REAL", "A", "Z"]

    real1_dates = provider.dates(resampling_frequency=None, realizations=[1])
    assert len(real1_dates) == 2
    date_to_get = real1_dates[0]
    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A", "Z"])
    assert vecdf.shape == (2, 3)
    assert vecdf.columns.tolist() == ["REAL", "A", "Z"]
    date_to_get = real1_dates[1]
    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A", "Z"])
    assert vecdf.shape == (1, 3)
    assert vecdf.columns.tolist() == ["REAL", "A", "Z"]



import inspect
from typing import Any, Callable, List, Optional, Union

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from pandas.api.types import is_numeric_dtype


def create_figure(plot_type: str, **kwargs: Any) -> go.Figure:
    """Create subplots for selected parameters"""

    if kwargs["data_frame"].empty:
        return empty_figure_layout()

    plotargs = set_default_args(**kwargs)

    fig: go.Figure = make_initial_figure(plot_type=plot_type, **plotargs)
    fig = update_xaxes(fig, plot_type=plot_type, **plotargs)
    fig = update_yaxes(fig, plot_type=plot_type, **plotargs)
    fig = update_layout(fig, **plotargs)
    fig = update_traces(fig, **plotargs)
    fig = for_each_annotation(fig, **plotargs)

    if plot_type == "distribution":
        fig = convert_violin_to_distribution_plot(fig, **plotargs)

    return fig


def set_default_args(**plotargs: Any) -> dict:

    plotargs["histnorm"] = plotargs.get("histnorm", "percent")
    plotargs["barmode"] = plotargs.get("barmode", "group")
    plotargs["opacity"] = plotargs.get("opacity", 0.7)

    if plotargs.get("facet_col") is not None:
        facet_cols = plotargs["data_frame"][plotargs["facet_col"]].nunique()
        plotargs.update(
            facet_col_wrap=min(
                min([x for x in range(100) if (x * (x + 1)) >= facet_cols]),
                20,
            ),
            facet_row_spacing=max((0.08 - (0.00071 * facet_cols)), 0.03),
            facet_col_spacing=max((0.06 - (0.00071 * facet_cols)), 0.03),
        )
        plotargs["custom_data"] = plotargs.get("custom_data", []) + [
            plotargs["facet_col"]
        ]
    return plotargs


def make_initial_figure(plot_type: str, **plotargs: Any) -> Callable:

    if plot_type == "distribution":
        plot_type = "violin"

    # pylint: disable=protected-access
    plotfunc = getattr(px._chart_types, plot_type)

    valid_plotargs = inspect.signature(plotfunc).parameters.keys()
    plotargs = {arg: value for arg, value in plotargs.items() if arg in valid_plotargs}

    return plotfunc(**plotargs)


def update_xaxes(figure: go.Figure, plot_type: str, **kwargs: Any) -> go.Figure:
    data_frame = kwargs["data_frame"]
    facet_col = kwargs.get("facet_col")
    return figure.update_xaxes(
        gridwidth=1,
        gridcolor="#ECECEC",
        showline=kwargs.get("framed", True),
        linewidth=2,
        linecolor="black",
        mirror=True,
        title=None
        if facet_col is not None or not isinstance(kwargs.get("x"), str)
        else kwargs.get("x"),
        showticklabels=(data_frame[facet_col].nunique() <= 100)
        if facet_col is not None
        else None,
        tickangle=0,
        tickfont_size=max((20 - (0.4 * data_frame[facet_col].nunique())), 10)
        if facet_col is not None
        else None,
        fixedrange=plot_type == "distribution",
    ).update_xaxes(**kwargs.get("xaxis", {}))


def update_yaxes(figure: go.Figure, plot_type: str, **kwargs: Any) -> go.Figure:
    return figure.update_yaxes(
        showline=kwargs.get("framed", True),
        linewidth=2,
        linecolor="black",
        mirror=True,
        gridwidth=1,
        gridcolor="#ECECEC",
        fixedrange=plot_type == "distribution",
        showticklabels=plot_type != "distribution",
    ).update_yaxes(**kwargs.get("yaxis", {}))


def update_layout(figure: go.Figure, **kwargs: Any) -> go.Figure:
    return figure.update_layout(plot_bgcolor="white", bargap=0).update_layout(
        **kwargs.get("layout", {})
    )


# pylint: disable=unnecessary-lambda
def update_traces(figure: go.Figure, **kwargs: Any) -> go.Figure:
    data_frame = kwargs["data_frame"]
    facet_col = kwargs.get("facet_col")
    return (
        figure.update_traces(
            marker_size=max((20 - (1.5 * data_frame[facet_col].nunique())), 5)
            if facet_col is not None
            else 20,
            selector=lambda t: t["type"] in ["scatter", "scattergl"],
        )
        .update_traces(textposition="inside", selector=dict(type="pie"))
        .for_each_trace(lambda t: set_marker_color(t))
        .for_each_trace(
            lambda t: t.update(
                xbins_size=(t["x"].max() - t["x"].min()) / kwargs.get("nbins", 15)
            )
            if is_numeric_dtype(t["x"])
            else None,
            selector=dict(type="histogram"),
        )
    )


def set_marker_color(trace: go) -> go:
    marker_attributes = trace.marker.to_plotly_json()
    if (
        "color" in marker_attributes
        and isinstance(trace.marker.color, str)
        and "#" in trace.marker.color
    ):
        opacity = marker_attributes.get(
            "opacity", 0.5 if trace.type in ["scatter", "scattergl"] else 0.7
        )
        trace.update(marker_line=dict(color=trace.marker.color, width=1))
        trace.update(marker_color=hex_to_rgb(trace.marker.color, opacity=opacity))
        trace.update(marker_opacity=1)
    return trace


def for_each_annotation(figure: go.Figure, **kwargs: Any) -> go.Figure:
    data_frame = kwargs["data_frame"]
    facet_col = kwargs.get("facet_col")
    return figure.for_each_annotation(
        lambda a: a.update(
            text=(a.text.split("=")[-1]),
            visible=data_frame[facet_col].nunique() <= 42
            if facet_col is not None
            else None,
            font_size=max((18 - (0.4 * data_frame[facet_col].nunique())), 10)
            if facet_col is not None
            else None,
        )
    )


def empty_figure_layout() -> go.Figure:
    return go.Figure(
        layout=dict(
            xaxis={"visible": False},
            yaxis={"visible": False},
            plot_bgcolor="white",
            annotations=[
                dict(
                    text="No data available for figure",
                    xref="paper",
                    yref="paper",
                    showarrow=False,
                    font={"size": 20},
                )
            ],
        )
    )


def convert_violin_to_distribution_plot(figure: go.Figure, **kwargs: Any) -> go.Figure:
    figure.for_each_trace(
        lambda t: t.update(
            y0=0,
            hoveron="violins",
            hoverinfo="none",
            meanline_visible=True,
            orientation="h",
            side="positive",
            width=2,
            points=False,
        )
    )
    hovertraces = create_hover_boxes_for_violin_plots(
        figure=figure,
        dframe=kwargs["data_frame"],
        facet_col=kwargs.get("facet_col", None),
        color_col=kwargs.get("color", None),
        x=kwargs["x"],
    )
    for trace in hovertraces:
        figure.add_trace(trace)
    return figure


def create_hover_boxes_for_violin_plots(
    figure: go.Figure,
    dframe: pd.DataFrame,
    x: str,
    facet_col: Optional[str],
    color_col: Optional[str],
) -> list:

    # Create invisible boxes used for hoverinfo on the violin plots
    # Necessary due to https://github.com/plotly/plotly.js/issues/2145
    hovertraces = []
    for trace in figure["data"]:
        hover_text = hover_box_text(trace, dframe, x, facet_col, color_col)
        hovertraces.append(
            go.Scatter(
                x=[min(trace.x), min(trace.x), max(trace.x), max(trace.x)],
                y=[0, 1, 1, 0],
                xaxis=trace.xaxis,
                yaxis=trace.yaxis,
                mode="lines",
                fill="toself",
                opacity=0,
                showlegend=False,
                text=hover_text,
                hoverinfo="text",
                hoverlabel=dict(bgcolor="#E6FAEC", font=dict(color="#243746", size=15)),
            )
        )
    return hovertraces


def hover_box_text(
    trace: dict,
    dframe: pd.DataFrame,
    x: str,
    facet_col: Optional[str],
    color_col: Optional[str],
) -> str:
    colors = list(dframe[color_col].unique()) if color_col in dframe else []
    facet = trace["customdata"][-1][0] if trace["customdata"] is not None else ""
    text = f"<b>{x}</b><br>" f"<b>{facet}</b><br>"
    if not colors:
        series = get_filtered_x_series(dframe, x, facet_col, facet)
        text = text + (
            f"Avg: {series.mean():{'.3g'}}<br>" f"Std: {series.std():{'.3g'}}<br>"
        )
    else:
        for color_item in colors:
            series = get_filtered_x_series(
                dframe, x, facet_col, facet, color_col, color_item
            )
            if series.empty:
                continue
            if color_item != facet:
                text = text + f"<b>{color_item}:</b><br>"
            text = text + (
                f"Avg: {series.mean():{'.3g'}}<br>" f"Std: {series.std():{'.3g'}}<br>"
            )
    return text


# pylint: disable=inconsistent-return-statements
def get_filtered_x_series(
    dframe: pd.DataFrame,
    x: str,
    facet_col: Optional[str] = None,
    facet: Optional[str] = None,
    color_col: Optional[str] = None,
    color: Optional[str] = None,
) -> pd.Series:
    if facet_col is None and color_col is None:
        return dframe[x]
    if facet_col is None and color_col is not None:
        return dframe.loc[dframe[color_col] == color][x]
    if facet_col is not None and color_col is None:
        return dframe.loc[dframe[facet_col] == facet][x]
    if facet_col is not None and color_col is not None:
        return dframe.loc[(dframe[facet_col] == facet) & (dframe[color_col] == color)][
            x
        ]


def hex_to_rgb(hex_string: str, opacity: float = 1) -> str:
    """Converts a hex color to rgb"""
    hex_string = hex_string.lstrip("#")
    hlen = len(hex_string)
    rgb: List[Union[int, float]] = [
        int(hex_string[i : i + hlen // 3], 16) for i in range(0, hlen, hlen // 3)
    ]
    rgb.append(opacity)
    return f"rgba{tuple(rgb)}"



from typing import Any, Dict, List, Optional, Sequence

import numpy as np
import pandas as pd

from ...._utils.colors import find_intermediate_color, rgba_to_str
from ...._utils.fanchart_plotting import (
    FanchartData,
    FreeLineData,
    LowHighData,
    MinMaxData,
    TraceDirection,
    get_fanchart_traces,
)
from .._types import DepthType, LineType
from ._rft_plotter_data_model import filter_frame, interpolate_depth


class FormationFigure:
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        well: str,
        ertdf: pd.DataFrame,
        enscolors: dict,
        depthtype: DepthType,
        date: str,
        ensembles: List[str],
        simdf: Optional[pd.DataFrame] = None,
        obsdf: Optional[pd.DataFrame] = None,
        reals: Optional[Sequence] = None,
    ) -> None:
        self._well = well
        rft_filter = {"WELL": well, "DATE": date, "ENSEMBLE": ensembles}
        if reals is not None:
            rft_filter.update({"REAL": reals})

        self._simdf = filter_frame(simdf, rft_filter) if simdf is not None else None
        self._obsdf = (
            filter_frame(obsdf, {"WELL": well, "DATE": date})
            if obsdf is not None
            else None
        )
        self._ertdf = filter_frame(ertdf, rft_filter)
        self._depthtype = depthtype
        self._enscolors = enscolors
        self._set_depth_columns()

        self._traces: List[Dict[str, Any]] = []
        self._layout = {
            "yaxis": {"autorange": "reversed", "title": "Depth", "showgrid": False},
            "xaxis": {
                "title": "Pressure",
                "showgrid": False,
                "range": [
                    (self.pressure_range[0] - self.xaxis_extension),
                    (self.pressure_range[1] + self.xaxis_extension),
                ],
            },
            "legend": {"orientation": "h"},
            "margin": {"t": 50},
            "hovermode": "closest",
            "uirevision": f"{well}{depthtype.value}",
            "title": f"{well} at {date}",
        }

    @property
    def figure(self) -> Dict[str, Any]:
        return {"data": self._traces, "layout": self._layout}

    @property
    def xaxis_extension(self) -> float:
        min_x, max_x = self.pressure_range
        return (max_x - min_x) * 0.5

    @property
    def pressure_range(self) -> List[float]:
        min_sim = (
            self._ertdf["SIMULATED"].min()
            if self.use_ertdf
            else self._simdf["PRESSURE"].min()
        )
        max_sim = (
            self._ertdf["SIMULATED"].max()
            if self.use_ertdf
            else self._simdf["PRESSURE"].max()
        )
        min_obs = (self._ertdf["OBSERVED"] - self._ertdf["OBSERVED_ERR"]).min()
        max_obs = (self._ertdf["OBSERVED"] + self._ertdf["OBSERVED_ERR"]).max()

        return [
            min(
                min_sim,
                min_obs,
            ),
            max(
                max_sim,
                max_obs,
            ),
        ]

    @property
    def simdf_has_md(self) -> bool:
        if (
            self._simdf is not None
            and "CONMD" in self._simdf
            and len(self._simdf["CONMD"].unique()) == len(self._simdf["DEPTH"].unique())
        ):
            return True
        return False

    @property
    def use_ertdf(self) -> bool:
        return self._simdf is None or (
            self._depthtype == DepthType.MD.value and not self.simdf_has_md
        )

    @property
    def ertdf_empty(self) -> bool:
        return self._ertdf.empty

    def _set_depth_columns(self) -> None:
        """Set depth columns (md vs tvd)"""
        self._ertdf["DEPTH"] = self._ertdf["TVD"]

        if self._depthtype == DepthType.MD:
            self._ertdf["DEPTH"] = self._ertdf[DepthType.MD.value]
            if self.simdf_has_md:
                self._simdf["DEPTH"] = self._simdf["CONMD"]
            if self._obsdf is not None and DepthType.MD.value in self._obsdf:
                self._obsdf["DEPTH"] = self._obsdf[DepthType.MD.value]

    def add_formation(self, df: pd.DataFrame, fill_color: bool = True) -> None:
        """Plot zonation"""
        formation_names = []
        formation_colors = [
            f"hsl({210*(1-x) + 202*x}, {100*(1-x) + 56*x}%, {98*(1-x) + 62*x}%)"
            for x in np.linspace(0, 1, len(df["ZONE"].unique()))
        ]

        top_col = "TOP_TVD" if self._depthtype == DepthType.TVD else "TOP_MD"
        base_col = "BASE_TVD" if self._depthtype == DepthType.TVD else "BASE_MD"

        df = filter_frame(df, {"WELL": self._well})
        df = df[df["TOP_MD"] != df["BASE_MD"]]

        for (_index, row) in df.iterrows():
            if pd.isnull(row[base_col]):
                raise KeyError(f"{base_col} has missing information")

            formation_names.append(
                {
                    "xref": "paper",
                    "yref": "y",
                    "x0": 0,
                    "x1": 1,
                    "name": row["ZONE"],
                    "y0": row[top_col],
                    "y1": row[base_col],
                    "line": dict(color="#646567", width=1.0),
                    "fillcolor": formation_colors[
                        list(df["ZONE"].unique()).index(row["ZONE"])
                    ]
                    if fill_color
                    else None,
                    "type": "rect",
                    "layer": "below",
                }
            )
            # Add formation names
            self._traces.append(
                {
                    "showlegend": _index - 1 == df.index.min(),
                    "name": "Formations",
                    "legendgroup": "name",
                    "type": "scatter",
                    "y": [abs((row[base_col] + row[top_col]) / 2)],
                    "x": [self.pressure_range[1] + (self.xaxis_extension / 1.2)],
                    "text": f"<b>{row['ZONE']}</b>",
                    "textfont": {"size": 14, "color": "#E50000"},
                    "textposition": "middle left",
                    "mode": "text",
                    "hoverinfo": "skip",
                }
            )
        self._layout.update({"shapes": formation_names})

    def add_additional_observations(self) -> None:
        if self._obsdf is not None:
            self._traces.append(
                {
                    "x": self._obsdf["PRESSURE"],
                    "y": self._obsdf["DEPTH"],
                    "type": "scatter",
                    "mode": "markers",
                    "name": "Observations",
                    "marker": {"color": "Black", "size": 20},
                }
            )

    def add_ert_observed(self) -> None:
        df = self._ertdf
        if not df.empty:
            df = filter_frame(
                df,
                {
                    "REAL": df["REAL"].unique()[0],
                    "ENSEMBLE": df["ENSEMBLE"].unique()[0],
                },
            )
            self._traces.append(
                {
                    "x": df["OBSERVED"],
                    "y": df["DEPTH"],
                    "type": "scatter",
                    "mode": "markers",
                    "name": "Ert observations",
                    "marker": {
                        "color": "#2584DE",
                        "size": 20,
                    },
                    "error_x": {
                        "type": "data",
                        "array": df["OBSERVED_ERR"],
                        "visible": True,
                        "thickness": 6,
                    },
                }
            )

    def add_simulated_lines(self, linetype: LineType) -> None:
        if self.use_ertdf:
            for ensemble, ensdf in self._ertdf.groupby("ENSEMBLE"):
                self._traces.append(
                    {
                        "x": ensdf["SIMULATED"],
                        "y": ensdf["DEPTH"],
                        "type": "scatter",
                        "mode": "markers",
                        "name": ensemble,
                        "marker": {
                            "color": self._enscolors[ensemble],
                            "size": 20,
                            "line": {"width": 1, "color": "grey"},
                        },
                    }
                )
        else:
            if linetype == LineType.REALIZATION:
                for ensemble, ensdf in self._simdf.groupby("ENSEMBLE"):
                    for i, (real, realdf) in enumerate(ensdf.groupby("REAL")):
                        self._traces.append(
                            {
                                "x": realdf["PRESSURE"],
                                "y": realdf["DEPTH"],
                                "hoverinfo": "y+x+text",
                                "hovertext": f"Realization: {real}, Ensemble: {ensemble}",
                                "type": "scatter",
                                "mode": "lines",
                                "line": {"color": self._enscolors[ensemble]},
                                "name": ensemble,
                                "showlegend": i == 0,
                                "legendgroup": ensemble,
                                "customdata": real,
                            }
                        )
            if linetype == LineType.FANCHART:
                for ensemble, ensdf in self._simdf.groupby("ENSEMBLE"):
                    quantiles = [10, 90]
                    dframe = (
                        interpolate_depth(ensdf).drop(columns=["REAL"]).groupby("DEPTH")
                    )
                    # Build a dictionary of dataframes to be concatenated
                    dframes = {}
                    dframes["mean"] = dframe.mean()
                    for quantile in quantiles:
                        quantile_str = "p" + str(quantile)
                        dframes[quantile_str] = dframe.quantile(q=quantile / 100.0)
                    dframes["maximum"] = dframe.max()
                    dframes["minimum"] = dframe.min()
                    self._traces.extend(
                        _get_fanchart_traces(
                            pd.concat(dframes, names=["STATISTIC"], sort=False)[
                                "PRESSURE"
                            ],
                            self._enscolors[ensemble],
                            ensemble,
                        )
                    )

    def color_by_param_value(self, df_norm: pd.DataFrame, selected_param: str) -> None:
        """This method colors the RFT traces according to the value of the input
        selected_param
        """
        for trace in self._traces:
            if "customdata" in trace:
                trace["line"]["color"] = set_real_color(
                    real_no=trace["customdata"], df_norm=df_norm
                )
                trace["hovertext"] = (
                    f"Real: {str(trace['customdata'])}, {selected_param}: "
                    f"{df_norm.loc[df_norm['REAL'] == trace['customdata']].iloc[0]['VALUE']}"
                )

        self._layout["title"] = {
            "text": f"{self._well} colored by {selected_param}",
        }


def set_real_color(df_norm: pd.DataFrame, real_no: str) -> str:
    """
    Return color for trace based on normalized parameter value.
    Midpoint for the colorscale is set on the average value
    Same function as in Parameter Analysis. Shout be generalized
    """
    red = rgba_to_str((255, 18, 67, 1))
    mid_color = rgba_to_str((220, 220, 220, 1))
    green = rgba_to_str((62, 208, 62, 1))

    mean = df_norm["VALUE_NORM"].mean()

    norm_value = df_norm.loc[df_norm["REAL"] == real_no].iloc[0]["VALUE_NORM"]
    if norm_value <= mean:
        intermed = norm_value / mean
        return find_intermediate_color(red, mid_color, intermed)

    intermed = (norm_value - mean) / (1 - mean)
    return find_intermediate_color(mid_color, green, intermed)


def _get_fanchart_traces(
    vector_stats: pd.DataFrame, color: str, legend_group: str
) -> List[Dict[str, Any]]:
    """Renders a fanchart for an ensemble vector"""

    # Retrieve indices from one of the keys
    x = vector_stats["maximum"].index.tolist()

    data = FanchartData(
        samples=x,
        low_high=LowHighData(
            low_data=vector_stats["p90"].values,
            low_name="P90",
            high_data=vector_stats["p10"].values,
            high_name="P10",
        ),
        minimum_maximum=MinMaxData(
            minimum=vector_stats["minimum"].values,
            maximum=vector_stats["maximum"].values,
        ),
        free_line=FreeLineData("Mean", vector_stats["mean"].values),
    )

    return get_fanchart_traces(
        data=data,
        hex_color=color,
        legend_group=legend_group,
        direction=TraceDirection.VERTICAL,
    )



import datetime
from typing import Optional

import numpy as np
import pandas as pd

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.dataframe_utils import (
    assert_date_column_is_datetime_object,
    make_date_column_datetime_object,
)

###################################################################################
# NOTE: This code is a copy and modification of
# webviz_subsurface/_datainput/from_timeseries_cumulatives.py for usage with
# EnsembleSummaryProvider functionality
#
# Renaming is performed for clarity
#
# Some additional functions are added.
###################################################################################


def is_per_interval_or_per_day_vector(vector: str) -> bool:
    return vector.startswith("PER_DAY_") or vector.startswith("PER_INTVL_")


def get_cumulative_vector_name(vector: str) -> str:
    if not is_per_interval_or_per_day_vector(vector):
        raise ValueError(
            f'Expected "{vector}" to be a vector calculated from cumulative!'
        )

    if vector.startswith("PER_DAY_"):
        return vector.lstrip("PER_DAY_")
    if vector.startswith("PER_INTVL_"):
        return vector.lstrip("PER_INTVL_")
    raise ValueError(f"Expected {vector} to be a cumulative vector!")


def create_per_day_vector_name(vector: str) -> str:
    return f"PER_DAY_{vector}"


def create_per_interval_vector_name(vector: str) -> str:
    return f"PER_INTVL_{vector}"


def calculate_from_resampled_cumulative_vectors_df(
    vectors_df: pd.DataFrame,
    as_per_day: bool,
) -> pd.DataFrame:
    """
    Calculates interval delta or average per day data for vector columns in provided dataframe.
    This function assumes data is already resampled when retrieved with ensemble summary
    provider.

    Date sampling should be according to Frequency enum in
    webviz_subsurface/_providers/ensemble_summary_provider.py.
    Does not handle raw data and varying time delta less than daily!

    `INPUT:`
    * vectors_df: pd.Dataframe - Dataframe with columns:
        ["DATE", "REAL", vector1, ..., vectorN]

    `NOTE:`
    - Does not handle raw data format with varying sampling or sampling frequency higher than
    daily!
    - Dataframe has columns:\n
    "DATE": Series with dates on datetime.datetime format
    "REAL": Series of realization number identifier
    vector1, ..., vectorN: Series of vector data for vector of given column name

    `TODO:`
    * IMPROVE FUNCTION NAME?
    * Handle raw data format?
    * Give e.g. a dict with info of "per_day and per_intvl" calculation for each vector column?
    Can thereby calculate everything for provided vector columns and no iterate column per
    column?
    """
    assert_date_column_is_datetime_object(vectors_df)

    vectors_df = vectors_df.copy()

    column_keys = [elm for elm in vectors_df.columns if elm not in ["DATE", "REAL"]]

    # Sort by realizations, thereafter dates
    vectors_df.sort_values(by=["REAL", "DATE"], inplace=True)

    # Create column of unique id for realizations. .diff() takes diff between an index
    # and previous index in a column. Thereby if "realuid" is != 0, the .diff() is
    # between two realizations.
    # Could alternatively loop over ensembles and realizations, but this is quicker for
    # larger datasets.
    vectors_df["realuid"] = vectors_df["REAL"]

    vectors_df.set_index(["DATE"], inplace=True)

    cumulative_name_map = {
        vector: create_per_day_vector_name(vector)
        if as_per_day
        else create_per_interval_vector_name(vector)
        for vector in column_keys
    }
    cumulative_vectors = list(cumulative_name_map.values())

    # Take diff of given column_keys indexes - preserve REAL
    cumulative_vectors_df = pd.concat(
        [
            vectors_df[["REAL"]],
            vectors_df[["realuid"] + column_keys]
            .diff()
            .shift(-1)
            .rename(
                mapper=cumulative_name_map,
                axis=1,
            ),
        ],
        axis=1,
    )
    cumulative_vectors_df[cumulative_vectors] = cumulative_vectors_df[
        cumulative_vectors
    ].fillna(value=0)

    # Reset index (DATE becomes regular column)
    cumulative_vectors_df.reset_index(inplace=True)

    # Convert interval cumulative to daily average rate if requested
    if as_per_day:
        days = cumulative_vectors_df["DATE"].diff().shift(-1).dt.days.fillna(value=0)
        for vector in column_keys:
            with np.errstate(invalid="ignore"):
                cumulative_vector_name = cumulative_name_map[vector]
                cumulative_vectors_df.loc[:, cumulative_vector_name] = (
                    cumulative_vectors_df[cumulative_vector_name].values / days.values
                )

    # Find .diff() between two realizations and set value = 0
    cumulative_vectors_df.loc[
        cumulative_vectors_df["realuid"] != 0, cumulative_vectors
    ] = 0
    cumulative_vectors_df.drop("realuid", axis=1, inplace=True)

    make_date_column_datetime_object(cumulative_vectors_df)

    return cumulative_vectors_df


# pylint: disable=too-many-return-statements
def datetime_to_intervalstr(date: datetime.datetime, freq: Frequency) -> Optional[str]:
    if date is None:
        return None

    if freq == Frequency.DAILY:
        return f"{date.year}-{date.month:02d}-{date.day:02d}"
    if freq == Frequency.WEEKLY:
        # Note: weekyear may differ from actual year for week numbers 1,52,53.
        (weekyear, week, _) = date.isocalendar()
        return f"{weekyear}-W{week:02d}"
    if freq == Frequency.MONTHLY:
        return f"{date.year}-{date.month:02d}"
    if freq == Frequency.QUARTERLY:
        return f"{date.year}-Q{1 + (date.month-1)//3}"
    if freq == Frequency.YEARLY:
        return f"{date.year}"

    # Using isoformat if frequency is none of the listed instead of error.
    return date.isoformat()



import logging
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from webviz_config.common_cache import CACHE

from webviz_subsurface._models import GruptreeModel
from webviz_subsurface._providers import EnsembleSummaryProvider

from .._types import DataType, EdgeOrNode, NodeType, StatOptions, TreeModeOptions


class EnsembleGroupTreeData:
    """This class holds the summary provider and gruptree dataset. It has functionality
    to combine the data and calculate the GroupTree component input dataset.
    """

    def __init__(
        self,
        provider: EnsembleSummaryProvider,
        gruptree_model: GruptreeModel,
        terminal_node: str,
        excl_well_startswith: Optional[List[str]] = None,
        excl_well_endswith: Optional[List[str]] = None,
    ):

        self._provider = provider
        self._gruptree_model = gruptree_model
        self._terminal_node = terminal_node
        self._gruptree = self._gruptree_model.get_filtered_dataframe(
            terminal_node=self._terminal_node,
            excl_well_startswith=excl_well_startswith,
            excl_well_endswith=excl_well_endswith,
        )

        self._wells: List[str] = self._gruptree[
            self._gruptree["KEYWORD"] == "WELSPECS"
        ]["CHILD"].unique()

        # Check that all WSTAT summary vectors exist
        # They are used to determine which summary vector are needed next.
        self._check_that_sumvecs_exists([f"WSTAT:{well}" for well in self._wells])

        # Add nodetypes IS_PROD, IS_INJ and IS_OTHER to gruptree
        self._gruptree = add_nodetype(
            self._gruptree, self._provider, self._wells, self._terminal_node
        )

        self._has_waterinj = False
        self._has_gasinj = False
        if True in self._gruptree["IS_INJ"].unique():
            # If there is injection in the tree we need to determine
            # which kind of injection. For that wee need FWIR and FGIR
            self._check_that_sumvecs_exists(["FWIR", "FGIR"])
            smry = self._provider.get_vectors_df(["FWIR", "FGIR"], None)
            self._has_waterinj = smry["FWIR"].sum() > 0
            self._has_gasinj = smry["FGIR"].sum() > 0

        # Add edge label
        self._gruptree["EDGE_LABEL"] = self._gruptree.apply(get_edge_label, axis=1)

        # Get summary data with metadata (nodename, datatype, edge_or_node)
        self._sumvecs: pd.DataFrame = self._get_sumvecs_with_metadata()

        # Check that all edge summary vectors exist
        self._check_that_sumvecs_exists(
            list(self._sumvecs[self._sumvecs["EDGE_NODE"] == EdgeOrNode.EDGE]["SUMVEC"])
        )

    @property
    def webviz_store(self) -> Tuple[Callable, List[Dict]]:
        return self._gruptree_model.webviz_store

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def create_grouptree_dataset(
        self,
        tree_mode: TreeModeOptions,
        stat_option: StatOptions,
        real: int,
        node_types: List[NodeType],
    ) -> Tuple[List[Dict[Any, Any]], List[Dict[str, str]], List[Dict[str, str]]]:
        """This method is called when an event is triggered to create a new dataset
        to the GroupTree plugin. First there is a lot of filtering of the smry and
        grouptree data, before the filtered data is sent to the function that is
        actually creating the dataset.

        Returns the group tree data and two lists with dropdown options for what
        to display on the edges and nodes.

        A sample data set can be found here:
        https://github.com/equinor/webviz-subsurface-components/blob/master/react/src/demo/example-data/group-tree.json
        """  # noqa

        # Filter smry
        vectors = [
            sumvec
            for sumvec in self._sumvecs["SUMVEC"]
            if sumvec in self._provider.vector_names()
        ]
        smry = self._provider.get_vectors_df(vectors, None)

        if tree_mode is TreeModeOptions.STATISTICS:
            if stat_option is StatOptions.MEAN:
                smry = smry.groupby("DATE").mean().reset_index()
            elif stat_option in [StatOptions.P50, StatOptions.P10, StatOptions.P90]:
                quantile = {"p50": 0.5, "p10": 0.9, "p90": 0.1}[stat_option.value]
                smry = smry.groupby("DATE").quantile(quantile).reset_index()
            elif stat_option is StatOptions.MAX:
                smry = smry.groupby("DATE").max().reset_index()
            elif stat_option is StatOptions.MIN:
                smry = smry.groupby("DATE").min().reset_index()
            else:
                raise ValueError(
                    f"Statistical option: {stat_option.value} not implemented"
                )
        else:
            smry = smry[smry["REAL"] == real]

        gruptree_filtered = self._gruptree
        if (
            tree_mode == TreeModeOptions.SINGLE_REAL
            and not self.tree_is_equivalent_in_all_real()
        ):
            # Trees are not equal. Filter on realization
            gruptree_filtered = gruptree_filtered[gruptree_filtered["REAL"] == real]

        # Filter nodetype prod, inj and/or other
        dfs = []
        for tpe in node_types:
            dfs.append(gruptree_filtered[gruptree_filtered[f"IS_{tpe.value}".upper()]])
        gruptree_filtered = pd.concat(dfs).drop_duplicates()

        return (
            create_dataset(smry, gruptree_filtered, self._sumvecs, self._terminal_node),
            self.get_edge_options(node_types),
            [
                {"name": datatype, "label": get_label(datatype)}
                for datatype in [DataType.PRESSURE, DataType.BHP, DataType.WMCTL]
            ],
        )

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def get_unique_real(self) -> List[int]:
        """Returns a list of unique realizations"""
        return self._provider.realizations()

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def tree_is_equivalent_in_all_real(self) -> bool:
        """Checks if the group tree is equivalent in all realizations,
        in which case there is only one REAL number in the dataframe
        """
        return self._gruptree["REAL"].nunique() == 1

    def _get_sumvecs_with_metadata(
        self,
    ) -> pd.DataFrame:
        """Returns a dataframe with the summary vectors that is needed to
        put together the group tree dataset. The other columns are metadata:

        * nodename: name in eclipse network
        * datatype: oilrate, gasrate, pressure etc
        * edge_node: whether the datatype is edge (f.ex rates) or node (f.ex pressure)

        Rates are not required for the terminal node since they will not be used.
        """
        records = []

        unique_nodes = self._gruptree.drop_duplicates(subset=["CHILD", "KEYWORD"])
        for _, noderow in unique_nodes.iterrows():
            nodename = noderow["CHILD"]
            keyword = noderow["KEYWORD"]

            datatypes = [DataType.PRESSURE]
            if noderow["IS_PROD"] and nodename != self._terminal_node:
                datatypes += [DataType.OILRATE, DataType.GASRATE, DataType.WATERRATE]
            if (
                noderow["IS_INJ"]
                and self._has_waterinj
                and nodename != self._terminal_node
            ):
                datatypes.append(DataType.WATERINJRATE)
            if (
                noderow["IS_INJ"]
                and self._has_gasinj
                and nodename != self._terminal_node
            ):
                datatypes.append(DataType.GASINJRATE)
            if keyword == "WELSPECS":
                datatypes += [DataType.BHP, DataType.WMCTL]

            for datatype in datatypes:
                records.append(
                    {
                        "NODENAME": nodename,
                        "DATATYPE": datatype,
                        "EDGE_NODE": get_edge_node(datatype),
                        "SUMVEC": get_sumvec(datatype, nodename, keyword),
                    }
                )
        return pd.DataFrame(records)

    def _check_that_sumvecs_exists(self, check_sumvecs: List[str]) -> None:
        """Takes in a list of summary vectors and checks if they are
        present in the summary dataset. If any are missing, a ValueError
        is raised with the list of all missing summary vectors.
        """
        missing_sumvecs = [
            sumvec
            for sumvec in check_sumvecs
            if sumvec not in self._provider.vector_names()
        ]
        if missing_sumvecs:
            str_missing_sumvecs = ", ".join(missing_sumvecs)
            raise ValueError(
                "Missing summary vectors for the GroupTree plugin: "
                f"{str_missing_sumvecs}."
            )

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def get_edge_options(self, node_types: List[NodeType]) -> List[Dict[str, str]]:
        """Returns a list with edge node options for the dropdown
        menu in the GroupTree component. The output list has the format:
        [
            {"name": DataType.OILRATE, "label": "Oil Rate"},
            {"name": DataType.GasRATE, "label": "Gas Rate"},
        ]
        """
        options = []
        if NodeType.PROD in node_types:
            for rate in [DataType.OILRATE, DataType.GASRATE, DataType.WATERRATE]:
                options.append({"name": rate, "label": get_label(rate)})
        if NodeType.INJ in node_types and self._has_waterinj:
            options.append(
                {
                    "name": DataType.WATERINJRATE,
                    "label": get_label(DataType.WATERINJRATE),
                }
            )
        if NodeType.INJ in node_types and self._has_gasinj:
            options.append(
                {"name": DataType.GASINJRATE, "label": get_label(DataType.GASINJRATE)}
            )
        if options:
            return options
        return [{"name": DataType.OILRATE, "label": get_label(DataType.OILRATE)}]


def get_edge_label(row: pd.Series) -> str:
    """Returns the edge label for a row in the grouptree dataframe"""
    if (
        "VFP_TABLE" not in row
        or row["VFP_TABLE"] in [None, 9999]
        or np.isnan(row["VFP_TABLE"])
    ):
        return ""
    vfp_nb = int(row["VFP_TABLE"])
    return f"VFP {vfp_nb}"


def get_label(datatype: DataType) -> str:
    """Returns a more readable label for the summary datatypes"""
    labels = {
        DataType.OILRATE: "Oil Rate",
        DataType.GASRATE: "Gas Rate",
        DataType.WATERRATE: "Water Rate",
        DataType.WATERINJRATE: "Water Inj Rate",
        DataType.GASINJRATE: "Gas Inj Rate",
        DataType.PRESSURE: "Pressure",
        DataType.BHP: "BHP",
        DataType.WMCTL: "WMCTL",
    }
    if datatype in labels:
        return labels[datatype]
    raise ValueError(f"Label for datatype {datatype.value} not implemented.")


def get_sumvec(
    datatype: DataType,
    nodename: str,
    keyword: str,
) -> str:
    """Returns the correct summary vector for a given
    * datatype: oilrate, gasrate etc
    * nodename: FIELD, well name or group name in Eclipse network
    * keyword: GRUPTREE, BRANPROP or WELSPECS
    """
    datatype_map = {
        "FIELD": {
            DataType.OILRATE: "FOPR",
            DataType.GASRATE: "FGPR",
            DataType.WATERRATE: "FWPR",
            DataType.WATERINJRATE: "FWIR",
            DataType.GASINJRATE: "FGIR",
            DataType.PRESSURE: "GPR",
        },
        "GRUPTREE": {
            DataType.OILRATE: "GOPR",
            DataType.GASRATE: "GGPR",
            DataType.WATERRATE: "GWPR",
            DataType.WATERINJRATE: "GWIR",
            DataType.GASINJRATE: "GGIR",
            DataType.PRESSURE: "GPR",
        },
        # BRANPROP can not be used for injection, but the nodes
        # might also be GNETINJE and could therefore have injection.
        "BRANPROP": {
            DataType.OILRATE: "GOPRNB",
            DataType.GASRATE: "GGPRNB",
            DataType.WATERRATE: "GWPRNB",
            DataType.PRESSURE: "GPR",
            DataType.WATERINJRATE: "GWIR",
            DataType.GASINJRATE: "GGIR",
        },
        "WELSPECS": {
            DataType.OILRATE: "WOPR",
            DataType.GASRATE: "WGPR",
            DataType.WATERRATE: "WWPR",
            DataType.WATERINJRATE: "WWIR",
            DataType.GASINJRATE: "WGIR",
            DataType.PRESSURE: "WTHP",
            DataType.BHP: "WBHP",
            DataType.WMCTL: "WMCTL",
        },
    }
    if nodename == "FIELD":
        datatype_ecl = datatype_map["FIELD"][datatype]
        if datatype == "pressure":
            return f"{datatype_ecl}:{nodename}"
        return datatype_ecl
    try:
        datatype_ecl = datatype_map[keyword][datatype]
    except KeyError as exc:
        error = (
            f"Summary vector not found for eclipse keyword: {keyword}, "
            f"data type: {datatype.value} and node name: {nodename}. "
        )
        raise KeyError(error) from exc
    return f"{datatype_ecl}:{nodename}"


def get_edge_node(datatype: DataType) -> EdgeOrNode:
    """Returns if a given datatype is edge (typically rates) or node (f.ex pressures)"""
    if datatype in [
        DataType.OILRATE,
        DataType.GASRATE,
        DataType.WATERRATE,
        DataType.WATERINJRATE,
        DataType.GASINJRATE,
    ]:
        return EdgeOrNode.EDGE
    if datatype in [DataType.PRESSURE, DataType.BHP, DataType.WMCTL]:
        return EdgeOrNode.NODE
    raise ValueError(f"Data type {datatype.value} not implemented.")


def create_dataset(
    smry: pd.DataFrame,
    gruptree: pd.DataFrame,
    sumvecs: pd.DataFrame,
    terminal_node: str,
) -> List[dict]:
    """The function puts together the GroupTree component input dataset.

    The gruptree dataframe includes complete networks for every time
    the tree changes (f.ex if a new well is defined). The function loops
    through the trees and puts together all the summary data that is valid for
    the time span where the tree is valid, along with the tree structure itself.
    """
    trees = []
    # loop trees
    for date, gruptree_date in gruptree.groupby("DATE"):
        next_date = gruptree[gruptree.DATE > date]["DATE"].min()
        if pd.isna(next_date):
            next_date = smry["DATE"].max()
        smry_in_datespan = smry[
            (smry["DATE"] >= date) & (smry["DATE"] < next_date)
        ].groupby("DATE")
        dates = list(smry_in_datespan.groups)
        if dates:
            trees.append(
                {
                    "dates": [date.strftime("%Y-%m-%d") for date in dates],
                    "tree": extract_tree(
                        gruptree_date, terminal_node, smry_in_datespan, dates, sumvecs
                    ),
                }
            )
        else:
            logging.getLogger(__name__).info(
                f"""No summary data found for gruptree between {date} and {next_date}"""
            )
    return trees


def extract_tree(
    gruptree: pd.DataFrame,
    nodename: str,
    smry_in_datespan: Dict[Any, pd.DataFrame],
    dates: list,
    sumvecs: pd.DataFrame,
) -> dict:
    """Extract the tree part of the GroupTree component dataset. This functions
    works recursively and is initially called with the terminal node of the tree
    (usually FIELD)
    """
    # pylint: disable=too-many-locals
    node_sumvecs = sumvecs[sumvecs["NODENAME"] == nodename]
    nodedict = get_nodedict(gruptree, nodename)

    result: dict = {
        "node_label": nodename,
        "node_type": "Well" if nodedict["KEYWORD"] == "WELSPECS" else "Group",
        "edge_label": nodedict["EDGE_LABEL"],
    }

    edges = node_sumvecs[node_sumvecs["EDGE_NODE"] == EdgeOrNode.EDGE].to_dict(
        "records"
    )
    nodes = node_sumvecs[node_sumvecs["EDGE_NODE"] == EdgeOrNode.NODE].to_dict(
        "records"
    )

    edge_data: Dict[str, List[float]] = {item["DATATYPE"]: [] for item in edges}
    node_data: Dict[str, List[float]] = {item["DATATYPE"]: [] for item in nodes}

    # Looping the dates only once is very important for the speed of this function
    for _, smry_at_date in smry_in_datespan:
        for item in edges:
            edge_data[item["DATATYPE"]].append(
                round(smry_at_date[item["SUMVEC"]].values[0], 2)
            )
        for item in nodes:
            try:
                node_data[item["DATATYPE"]].append(
                    round(smry_at_date[item["SUMVEC"]].values[0], 2)
                )
            except KeyError:
                node_data[item["DATATYPE"]].append(np.nan)

    result["edge_data"] = edge_data
    result["node_data"] = node_data

    children = list(gruptree[gruptree["PARENT"] == nodename]["CHILD"].unique())
    if children:
        result["children"] = [
            extract_tree(gruptree, child, smry_in_datespan, dates, sumvecs)
            for child in children
        ]
    return result


def get_nodedict(gruptree: pd.DataFrame, nodename: str) -> Dict[str, Any]:
    """Returns the node data from a row in the gruptree dataframe as a dictionary.
    This function also checks that there is exactly one element with the given name.
    """
    df = gruptree[gruptree["CHILD"] == nodename]
    if df.empty:
        raise ValueError(f"No gruptree row found for node {nodename}")
    if df.shape[0] > 1:
        raise ValueError(f"Multiple gruptree rows found for node {nodename}. {df}")
    return df.to_dict("records")[0]


def add_nodetype(
    gruptree: pd.DataFrame,
    provider: EnsembleSummaryProvider,
    all_wells: List[str],
    terminal_node: str,
) -> pd.DataFrame:
    """Adds nodetype IS_PROD, IS_INJ and IS_OTHER."""

    # Get all nodes
    nodes = gruptree.drop_duplicates(subset=["CHILD"], keep="first").copy()

    # Identify leaf nodes (group nodes can also be leaf nodes)
    def is_leafnode(node: pd.Series) -> bool:
        if nodes[nodes["PARENT"] == node["CHILD"]].empty:
            return True
        return False

    nodes["IS_LEAF"] = nodes.apply(is_leafnode, axis=1)

    # Classify leaf nodes as producer, injector or other
    is_prod_map, is_inj_map, is_other_map = create_leafnodetype_maps(
        nodes[nodes["IS_LEAF"]], provider, all_wells
    )
    nodes["IS_PROD"] = nodes["CHILD"].map(is_prod_map)
    nodes["IS_INJ"] = nodes["CHILD"].map(is_inj_map)
    nodes["IS_OTHER"] = nodes["CHILD"].map(is_other_map)

    # Recursively find well types of all leaf nodes connected to the group node
    # Deduce group node type from well types
    nonleafs = nodes[~nodes["IS_LEAF"]]
    for _, node in nonleafs.iterrows():
        leafs_are_prod, leafs_are_inj, leafs_are_other = get_leafnode_types(
            node["CHILD"], nodes
        )
        is_prod_map[node["CHILD"]] = any(leafs_are_prod)
        is_inj_map[node["CHILD"]] = any(leafs_are_inj)
        is_other_map[node["CHILD"]] = any(leafs_are_other)

    # The terminal node must not be filtered out,
    # so it is set True for all categories
    is_prod_map[terminal_node] = True
    is_inj_map[terminal_node] = True
    is_other_map[terminal_node] = True

    # Tag all nodes as IS_PROD, IS_INJ and IS_OTHER
    gruptree["IS_PROD"] = gruptree["CHILD"].map(is_prod_map)
    gruptree["IS_INJ"] = gruptree["CHILD"].map(is_inj_map)
    gruptree["IS_OTHER"] = gruptree["CHILD"].map(is_other_map)
    return gruptree


def get_leafnode_types(
    node_name: str, gruptree: pd.DataFrame
) -> Tuple[List[Any], List[Any], List[Any]]:
    """This function finds the IS_PROD, IS_INJ and IS_OTHER values of all
    leaf nodes connected to the input node.

    The function is using recursion to find all wells below the node
    int the three.
    """
    children = gruptree[gruptree["PARENT"] == node_name]
    leafs_are_prod, leafs_are_inj, leafs_are_other = [], [], []
    for _, childrow in children.iterrows():
        if childrow["IS_LEAF"]:
            leafs_are_prod.append(childrow["IS_PROD"])
            leafs_are_inj.append(childrow["IS_INJ"])
            leafs_are_other.append(childrow["IS_OTHER"])
        else:
            prod, inj, other = get_leafnode_types(childrow["CHILD"], gruptree)
            leafs_are_prod += prod
            leafs_are_inj += inj
            leafs_are_other += other
    return leafs_are_prod, leafs_are_inj, leafs_are_other


def create_leafnodetype_maps(
    leafnodes: pd.DataFrame, provider: EnsembleSummaryProvider, all_wells: List[str]
) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
    """Returns three dictionaries classifying leaf nodes as producer,
    injector and/or other (f.ex observation well).

    Well leaf nodes are classified using WSTAT and group leaf nodes
    are classified using summary data.
    """
    # pylint: disable=too-many-locals
    is_prod_map, is_inj_map, is_other_map = {}, {}, {}
    wstat_df = provider.get_vectors_df([f"WSTAT:{well}" for well in all_wells], None)

    for _, leafnode in leafnodes.iterrows():
        nodename = leafnode["CHILD"]
        nodekeyword = leafnode["KEYWORD"]

        if nodekeyword == "WELSPECS":
            # The leaf node is a well
            wstat = wstat_df[f"WSTAT:{nodename}"].unique()
            is_prod_map[nodename] = 1 in wstat
            is_inj_map[nodename] = 2 in wstat
            is_other_map[nodename] = (1 not in wstat) and (2 not in wstat)
        else:
            # The leaf node is a group
            prod_sumvecs = [
                get_sumvec(datatype, nodename, nodekeyword)
                for datatype in [DataType.OILRATE, DataType.GASRATE, DataType.WATERRATE]
            ]
            inj_sumvecs = (
                [
                    get_sumvec(datatype, nodename, nodekeyword)
                    for datatype in [DataType.WATERINJRATE, DataType.GASINJRATE]
                ]
                if nodekeyword != "BRANPROP"
                else []
            )

            smry = provider.get_vectors_df(
                [
                    sumvec
                    for sumvec in (prod_sumvecs + inj_sumvecs)
                    if sumvec in provider.vector_names()
                ],
                None,
            )

            sumprod = sum(
                [
                    smry[sumvec].sum()
                    for sumvec in prod_sumvecs
                    if sumvec in smry.columns
                ]
            )

            suminj = sum(
                [smry[sumvec].sum() for sumvec in inj_sumvecs if sumvec in smry.columns]
            )

            is_prod_map[nodename] = sumprod > 0
            is_inj_map[nodename] = suminj > 0
            is_other_map[nodename] = (sumprod == 0) and (suminj == 0)
    return is_prod_map, is_inj_map, is_other_map



from typing import List

import pandas as pd
import webviz_core_components as wcc
from dash import html


def fipfile_qc_main_layout(uuid: str) -> wcc.Frame:
    return html.Div(
        children=[
            html.Div(
                style={"margin-bottom": "20px"},
                children=wcc.RadioItems(
                    vertical=False,
                    id={"id": uuid, "element": "display-option"},
                    options=[
                        {
                            "label": "QC plots",
                            "value": "plots",
                        },
                        {"label": "Table", "value": "table"},
                    ],
                    value="plots",
                ),
            ),
            html.Div(id=uuid),
        ],
    )


def fipfile_qc_selections(
    uuid: str,
    tab: str,
) -> wcc.Selectors:
    return wcc.Selectors(
        label="TABLE COTROLS",
        open_details=True,
        children=wcc.Checklist(
            id={"id": uuid, "tab": tab, "selector": "Group table"},
            options=[
                {"label": "Group table on set", "value": "grouped"},
            ],
            value=["grouped"],
        ),
    )


def fipfile_qc_filters(
    uuid: str, tab: str, disjoint_set_df: pd.DataFrame
) -> wcc.Selectors:
    return wcc.Selectors(
        label="FILTERS",
        open_details=True,
        children=filter_dropdowns(
            uuid=uuid,
            tab=tab,
            disjoint_set_df=disjoint_set_df,
        ),
    )


def filter_dropdowns(uuid: str, disjoint_set_df: pd.DataFrame, tab: str) -> html.Div:
    dropdowns: List[html.Div] = []
    for selector in ["REGION", "ZONE", "FIPNUM", "SET"]:
        elements = list(disjoint_set_df[selector].unique())
        if selector == "FIPNUM":
            elements = sorted(elements, key=int)
        dropdowns.append(
            html.Div(
                children=wcc.SelectWithLabel(
                    label=selector.lower().capitalize(),
                    id={"id": uuid, "tab": tab, "selector": selector, "type": "fipqc"},
                    options=[{"label": i, "value": i} for i in elements],
                    value=elements,
                    multi=True,
                    size=min(15, len(elements)),
                ),
            )
        )
    return html.Div(dropdowns)



import glob
import json
import logging
import re
from pathlib import Path
from typing import Any, Dict, List, Optional

import pandas as pd

# opm and ecl2df are only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from ecl2df import EclFiles, common
except ImportError:
    pass

from .fmu_input import scratch_ensemble


def remove_invalid_colors(zonelist: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Removes colors in the zonelist from the lyr file that is not 6 digit
    hexadecimal.
    """
    # pylint: disable=logging-fstring-interpolation
    for zonedict in zonelist:
        if "color" in zonedict and not re.match(
            "^#([A-Fa-f0-9]{6})", zonedict["color"]
        ):
            logging.getLogger(__name__).warning(
                f"""The zone color {zonedict["color"]} will be ignored. """
                "Only 6 digit hexadecimal colors are accepted in the well completions plugin."
            )
            zonedict.pop("color")
    return zonelist


def read_zone_layer_mapping(
    ensemble_path: str, zone_layer_mapping_file: str
) -> pd.DataFrame:
    """Searches for all zone->layer mapping files for an ensemble. \
    The files should be on lyr format and can be parsed using functionality \
    from ecl2df.

    The results are returned as a dataframe with the following columns:
    * REAL
    * K1 (layer)
    * ZONE
    * COLOR
    """
    ens = scratch_ensemble("ens", ensemble_path, filter_file="OK")
    df_files = ens.find_files(zone_layer_mapping_file)
    dataframes = []

    if df_files.empty:
        return pd.DataFrame()

    for _, row in df_files.iterrows():
        real = row["REAL"]
        zonelist = common.parse_lyrfile(filename=row["FULLPATH"])
        layer_zone_mapping = common.convert_lyrlist_to_zonemap(zonelist)
        df_real = pd.DataFrame.from_dict(
            layer_zone_mapping, orient="index", columns=["ZONE"]
        )
        df_real["K1"] = df_real.index
        df_real["REAL"] = real
        zonelist = remove_invalid_colors(zonelist)
        zone_color_mapping = {
            zonedict["name"]: zonedict["color"]
            for zonedict in zonelist
            if "color" in zonedict
        }
        df_real["COLOR"] = df_real["ZONE"].map(zone_color_mapping)
        dataframes.append(df_real)

    return pd.concat(dataframes)


def read_well_attributes(
    ensemble_path: str, well_attributes_file: str
) -> Optional[dict]:
    """Searches for a well attributes json file on the scratch disk. \
    if one file is found it is parsed and returned as a dictionary.

    The file needs to follow the format below. The categorical attributes \
    are optional.
    {
        "version" : "0.1",
        "wells" : [
        {
            "alias" : {
                "eclipse" : "OP_1"
            },
            "attributes" : {
                "mlt_singlebranch" : "mlt",
                "structure" : "East",
                "welltype" : "producer"
            },
            "name" : "OP_1"
        },
        {
            "alias" : {
                "eclipse" : "GI_1"
            },
            "attributes" : {
                "mlt_singlebranch" : "singlebranch",
                "structure" : "West",
                "welltype" : "gas injector"
            },
            "name" : "GI_1"
        },
        ]
    }
    """
    for filename in glob.glob(f"{ensemble_path}/{well_attributes_file}"):
        file_content = json.loads(Path(filename).read_text())
        return {
            well_data["alias"]["eclipse"]: well_data["attributes"]
            for well_data in file_content["wells"]
        }
    return None


def read_stratigraphy(
    ensemble_path: str, stratigraphy_file: str
) -> Optional[List[Dict]]:
    """Searches for a stratigraphy json file on the scratch disk. \
    If a file is found the content is returned as a list of dicts.
    """
    for filename in glob.glob(f"{ensemble_path}/{stratigraphy_file}"):
        return json.loads(Path(filename).read_text())
    return None


def get_ecl_datafile(ensemble_path: str) -> Optional[str]:
    """Returns the first eclipse DATA file found in the ensemble_path.
    If no DATA file is found it returns None.
    """
    for filename in glob.glob(f"{ensemble_path}/eclipse/model/*.DATA"):
        return filename
    return None


def get_ecl_unit_system(ensemble_path: str) -> Optional[str]:
    """Returns the unit system of an eclipse deck. The options are \
    METRIC, FIELD, LAB and PVT-M.

    If none of these are found, the function returns METRIC

    If no eclipse DATA file is found in the ensemble path, it
    returns None
    """
    datafile = get_ecl_datafile(ensemble_path)
    if datafile is None:
        return None
    ecl_deck = EclFiles(datafile).get_ecldeck()
    for keyword in ecl_deck:
        if keyword.name in ["METRIC", "FIELD", "LAB", "PVT-M"]:
            return keyword.name
    return "METRIC"


def read_well_connection_status(
    ensemble_path: str, well_connection_status_file: str
) -> Optional[pd.DataFrame]:
    """Reads csv file with well connection status data from the scratch disk.
    Merges together files from all realizations, does some fixing of the column
    data types, and returns it as a pandas dataframe.

    fmu-ensemble is used to find the file names on the scratch disk

    The well connection status data is extracted from the CPI data, which is 0 if the
    connection is SHUT and >0 if the connection is OPEN. This is independent of
    the status of the well.
    """

    ens = scratch_ensemble("ens", ensemble_path, filter_file="OK")
    df_files = ens.find_files(well_connection_status_file)

    if df_files.empty:
        return None

    df = pd.DataFrame()
    for _, row in df_files.iterrows():
        df_real = pd.read_csv(row.FULLPATH)
        df_real["REAL"] = row.REAL
        df = pd.concat([df, df_real])
    df.I = pd.to_numeric(df.I)
    df.J = pd.to_numeric(df.J)
    df["K1"] = pd.to_numeric(df.K)
    df = df.drop(["K"], axis=1)
    df.DATE = pd.to_datetime(df.DATE).dt.date
    return df



from typing import Callable, Dict, List, Optional, Union

from dash import ALL, Input, Output, State, callback, callback_context
from dash.exceptions import PreventUpdate

from ._business_logic import SwatinitQcDataModel
from ._figures import MapFigure, PropertiesVsDepthSubplots, WaterfallPlot
from ._layout import LayoutElements, TabMaxPcInfoLayout, TabQqPlotLayout, Tabs


def plugin_callbacks(get_uuid: Callable, datamodel: SwatinitQcDataModel) -> None:
    qc_plot_layout = TabQqPlotLayout(get_uuid, datamodel)
    table_layout = TabMaxPcInfoLayout(get_uuid, datamodel)

    @callback(
        Output(get_uuid(LayoutElements.PLOT_WRAPPER), "children"),
        Input(get_uuid(LayoutElements.SELECTED_TAB), "value"),
        Input(get_uuid(LayoutElements.PLOT_EQLNUM_SELECTOR), "value"),
        Input({"id": get_uuid(LayoutElements.FILTERS_DISCRETE), "col": ALL}, "value"),
        Input({"id": get_uuid(LayoutElements.FILTERS_CONTINOUS), "col": ALL}, "value"),
        Input(get_uuid(LayoutElements.COLOR_BY), "value"),
        Input(get_uuid(LayoutElements.MAX_POINTS), "value"),
        Input(get_uuid(LayoutElements.PLOT_SELECTOR), "value"),
        State({"id": get_uuid(LayoutElements.FILTERS_DISCRETE), "col": ALL}, "id"),
        State({"id": get_uuid(LayoutElements.FILTERS_CONTINOUS), "col": ALL}, "id"),
    )
    # pylint: disable=too-many-arguments
    def _update_plot(
        tab_selected: str,
        eqlnums: List[str],
        dicrete_filters: List[List[str]],
        continous_filters: List[List[str]],
        color_by: str,
        max_points: int,
        plot_selector: str,
        dicrete_filters_ids: List[Dict[str, str]],
        continous_filters_ids: List[Dict[str, str]],
    ) -> list:

        if tab_selected != Tabs.QC_PLOTS or max_points is None:
            raise PreventUpdate

        filters = zip_filters(dicrete_filters, dicrete_filters_ids)
        filters.update({"EQLNUM": eqlnums})

        df = datamodel.get_dataframe(
            filters=filters,
            range_filters=zip_filters(continous_filters, continous_filters_ids),
        )
        if df.empty:
            return ["No data left after filtering"]

        qc_volumes = datamodel.compute_qc_volumes(df)

        df = datamodel.filter_dframe_on_depth(df)
        df = datamodel.resample_dataframe(df, max_points=max_points)

        colormap = datamodel.create_colormap(color_by)
        main_plot = (
            WaterfallPlot(qc_vols=qc_volumes).figure
            if plot_selector == qc_plot_layout.MainPlots.WATERFALL
            else PropertiesVsDepthSubplots(
                dframe=df,
                color_by=color_by,
                colormap=colormap,
                discrete_color=color_by in datamodel.SELECTORS,
            ).figure
        )
        map_figure = MapFigure(
            dframe=df,
            color_by=color_by,
            faultlinedf=datamodel.faultlines_df,
            colormap=colormap,
        ).figure

        return qc_plot_layout.main_layout(
            main_figure=main_plot,
            map_figure=map_figure,
            qc_volumes=qc_volumes,
        )

    @callback(
        Output(get_uuid(LayoutElements.MAIN_FIGURE), "figure"),
        Output(get_uuid(LayoutElements.MAP_FIGURE), "figure"),
        Input(get_uuid(LayoutElements.MAIN_FIGURE), "selectedData"),
        Input(get_uuid(LayoutElements.MAP_FIGURE), "selectedData"),
        State(get_uuid(LayoutElements.MAIN_FIGURE), "figure"),
        State(get_uuid(LayoutElements.MAP_FIGURE), "figure"),
    )
    def _update_selected_points_in_figures(
        selected_data: dict, selected_data_map: dict, mainfig: dict, mapfig: dict
    ) -> tuple:
        ctx = callback_context.triggered[0]["prop_id"]

        selected = (
            selected_data_map if LayoutElements.MAP_FIGURE in ctx else selected_data
        )
        point_indexes = get_point_indexes_from_selected(selected)

        for trace in mainfig["data"]:
            update_selected_points_in_trace(trace, point_indexes)
        for trace in mapfig["data"]:
            update_selected_points_in_trace(trace, point_indexes)

        return mainfig, mapfig

    def get_point_indexes_from_selected(selected: Optional[dict]) -> Union[list, dict]:
        if not (isinstance(selected, dict) and "points" in selected):
            return []

        continous_color = "marker.color" in selected["points"][0]
        if continous_color:
            return [point["pointNumber"] for point in selected["points"]]

        point_indexes: dict = {}
        for point in selected["points"]:
            trace_name = str(point["customdata"][0])
            if trace_name not in point_indexes:
                point_indexes[trace_name] = []
            point_indexes[trace_name].append(point["pointNumber"])
        return point_indexes

    def update_selected_points_in_trace(
        trace: dict, point_indexes: Union[dict, list]
    ) -> None:
        if "name" in trace:
            selectedpoints = (
                point_indexes
                if isinstance(point_indexes, list)
                else point_indexes.get(trace["name"], [])
            )
            trace.update(selectedpoints=selectedpoints if point_indexes else None)

    @callback(
        Output(get_uuid(LayoutElements.TABLE_WRAPPER), "children"),
        Input(get_uuid(LayoutElements.SELECTED_TAB), "value"),
        Input(get_uuid(LayoutElements.TABLE_EQLNUM_SELECTOR), "value"),
        Input(get_uuid(LayoutElements.HIGHLIGHT_ABOVE), "value"),
        Input(get_uuid(LayoutElements.GROUPBY_EQLNUM), "value"),
        Input(
            {"id": get_uuid(LayoutElements.FILTERS_CONTINOUS_MAX_PC), "col": ALL},
            "value",
        ),
        State(
            {"id": get_uuid(LayoutElements.FILTERS_CONTINOUS_MAX_PC), "col": ALL}, "id"
        ),
    )
    def _update_capillary_pressure_tab_layout(
        tab_selected: str,
        eqlnums: list,
        threshold: Optional[int],
        groupby_eqlnum: list,
        continous_filters: List[List[str]],
        continous_filters_ids: List[Dict[str, str]],
    ) -> str:
        if tab_selected != Tabs.MAX_PC_SCALING:
            raise PreventUpdate

        df = datamodel.get_dataframe(
            filters={"EQLNUM": eqlnums},
            range_filters=zip_filters(continous_filters, continous_filters_ids),
        )
        df_for_map = df[df["PC_SCALING"] >= threshold]
        if threshold is None:
            df_for_map = datamodel.resample_dataframe(df, max_points=10000)

        map_figure = MapFigure(
            dframe=df_for_map,
            color_by="EQLNUM",
            faultlinedf=datamodel.faultlines_df,
            colormap=datamodel.create_colormap("EQLNUM"),
        ).figure
        df = datamodel.get_max_pc_info_and_percent_for_data_matching_condition(
            dframe=df,
            condition=threshold,
            groupby_eqlnum=groupby_eqlnum == "both",
        )
        return table_layout.main_layout(
            dframe=df,
            selectors=datamodel.SELECTORS,
            map_figure=map_figure,
        )

    @callback(
        Output(get_uuid("info-dialog"), "open"),
        Input(get_uuid("info-button"), "n_clicks"),
        State(get_uuid("info-dialog"), "open"),
    )
    def open_close_information_dialog(_n_click: list, is_open: bool) -> bool:
        if _n_click is not None:
            return not is_open
        raise PreventUpdate


def zip_filters(filter_values: list, filter_ids: list) -> dict:
    return {id_val["col"]: values for values, id_val in zip(filter_values, filter_ids)}



import datetime
from typing import Dict

import dateutil.parser  # type: ignore
import numpy as np
import pandas as pd


def make_date_column_datetime_object(df: pd.DataFrame) -> pd.DataFrame:

    # Make a copy since it is likely we will modify the dataframe, and
    # we don't know if it is a view on a larger DF or a copy
    df = df.copy()

    if "DATE" not in df.columns:
        return df

    sampled_date_value = df["DATE"].values[0]

    # Infer datatype (Pandas cannot answer it) based on the first element:
    if isinstance(sampled_date_value, pd.Timestamp):
        df["DATE"] = pd.Series(pd.to_pydatetime(df["DATE"]), dtype="object")

    elif isinstance(sampled_date_value, str):
        # Do not use pd.Series.apply() here, Pandas would try to convert it to
        # datetime64[ns] which is limited at year 2262.
        df["DATE"] = pd.Series(
            [dateutil.parser.parse(datestr) for datestr in df["DATE"]], dtype="object"
        )

    elif isinstance(sampled_date_value, datetime.date):
        df["DATE"] = pd.Series(
            [
                datetime.datetime.combine(dateobj, datetime.datetime.min.time())
                for dateobj in df["DATE"]
            ],
            dtype="object",
        )

    return df


def find_min_max_for_numeric_df_columns(
    df: pd.DataFrame,
) -> Dict[str, dict]:

    ret_dict = {}
    for col_name in df.columns:
        series = df[col_name]
        if pd.api.types.is_numeric_dtype(series):
            # minval = series.min().item()
            # maxval = series.max().item()

            nparr = series.values
            minval = np.nanmin(nparr).item()
            maxval = np.nanmax(nparr).item()

            ret_dict[col_name] = {"min": minval, "max": maxval}

    return ret_dict

    # Using df.describe() seems to be very slow!
    # desc_df = df.describe(percentiles=[], include=[np.number])
    # ret_dict = {}
    # for vec_name in desc_df.columns:
    #     minval = desc_df[vec_name]["min"]
    #     maxval = desc_df[vec_name]["max"]
    #     ret_dict[vec_name] = {"min": minval, "max": maxval}

    # return ret_dict



import json
import threading
from typing import Dict, Optional, Union

from .ensemble_set_model import EnsembleSetModel

# Module level globals
# Do we actually need to consider locking here or will this access always be single threaded?
_cache_lock = threading.Lock()
_ensemble_set_model_cache: Dict[str, EnsembleSetModel] = {}


def get_or_create_model(
    ensemble_paths: dict,
    time_index: Optional[Union[list, str]] = None,
    column_keys: Optional[list] = None,
) -> EnsembleSetModel:

    modelkey = json.dumps(
        {
            "ensemble_paths": ensemble_paths,
            "time_index": time_index,
            "column_keys": column_keys,
        }
    )

    with _cache_lock:
        if modelkey in _ensemble_set_model_cache:
            # Just return existing model from cache
            return _ensemble_set_model_cache[modelkey]

        # No matching model in cache -> create a new ensemble set model and insert in cache
        new_model = EnsembleSetModel(
            ensemble_paths=ensemble_paths,
            smry_time_index=time_index,
            smry_column_keys=column_keys,
        )
        _ensemble_set_model_cache[modelkey] = new_model

        return new_model



from datetime import datetime
from pathlib import Path
from typing import Dict

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pytest

from webviz_subsurface._providers.ensemble_summary_provider._provider_impl_arrow_lazy import (
    Frequency,
    ProviderImplArrowLazy,
    _find_first_non_increasing_date_pair,
    _is_date_column_monotonically_increasing,
)
from webviz_subsurface._providers.ensemble_summary_provider.ensemble_summary_provider import (
    EnsembleSummaryProvider,
)


def _add_mock_smry_meta_to_table(table: pa.Table) -> pa.Table:
    schema = table.schema
    for colname in schema.names:
        is_rate = bool("_r" in colname)
        is_total = bool("_t" in colname)

        metadata = None
        if is_rate or is_total:
            metadata = {
                b"unit": b"N/A",
                b"is_rate": b"True" if is_rate else b"False",
                b"is_total": b"True" if is_total else b"False",
                b"is_historical": b"False",
                b"keyword": b"UNKNOWN",
            }

        if metadata:
            idx = schema.get_field_index(colname)
            field = schema.field(idx)
            field = field.with_metadata(metadata)
            schema = schema.set(idx, field)

    table = table.cast(schema)
    return table


def _split_into_per_realization_tables(table: pa.Table) -> Dict[int, pa.Table]:
    per_real_tables: Dict[int, pa.Table] = {}
    unique_reals = table.column("REAL").unique().to_pylist()
    for real in unique_reals:
        # pylint: disable=no-member
        mask = pc.is_in(table["REAL"], value_set=pa.array([real]))
        real_table = table.filter(mask).drop(["REAL"])
        per_real_tables[real] = real_table

    return per_real_tables


def _create_provider_obj_with_data(
    input_data: list,
    storage_dir: Path,
) -> EnsembleSummaryProvider:

    # Turn rows into columns
    columns_with_header = list(zip(*input_data))

    input_dict = {}
    for col in columns_with_header:
        colname = col[0]
        coldata = col[1:]
        input_dict[colname] = coldata
    input_table = pa.Table.from_pydict(input_dict)

    input_table = _add_mock_smry_meta_to_table(input_table)

    # Split into per realization tables
    per_real_tables = _split_into_per_realization_tables(input_table)

    ProviderImplArrowLazy.write_backing_store_from_per_realization_tables(
        storage_dir, "dummy_key", per_real_tables
    )
    new_provider = ProviderImplArrowLazy.from_backing_store(storage_dir, "dummy_key")

    if not new_provider:
        raise ValueError("Failed to create EnsembleSummaryProvider")

    return new_provider


def test_create_with_dates_after_2262(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                                 "REAL", "A"],
        [np.datetime64("2000-01-02T00:00", "ms"), 1,     10.0],
        [np.datetime64("2500-12-20T23:59", "ms"), 1,     12.0],
        [np.datetime64("2500-12-21T22:58", "ms"), 1,     13.0],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    dates = provider.dates(resampling_frequency=None)
    assert len(dates) == 3
    assert dates[0] == datetime(2000, 1, 2, 00, 00)
    assert dates[1] == datetime(2500, 12, 20, 23, 59)
    assert dates[2] == datetime(2500, 12, 21, 22, 58)


def test_get_vector_names(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "A",  "C",   "Z"],
        [np.datetime64("2023-12-20", "ms"),  0,      10.0,  1.0,  0.0 ],
        [np.datetime64("2023-12-20", "ms"),  1,      12.0,  1.0,  0.0 ],
        [np.datetime64("2023-12-21", "ms"),  1,      13.0,  1.0,  0.0 ],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    all_vecnames = provider.vector_names()
    assert len(all_vecnames) == 3
    assert all_vecnames == ["A", "C", "Z"]

    non_const_vec_names = provider.vector_names_filtered_by_value(
        exclude_constant_values=True
    )
    assert len(non_const_vec_names) == 1
    assert non_const_vec_names == ["A"]

    non_zero_vec_names = provider.vector_names_filtered_by_value(
        exclude_all_values_zero=True
    )
    assert len(non_zero_vec_names) == 2
    assert non_zero_vec_names == ["A", "C"]

    all_realizations = provider.realizations()
    assert len(all_realizations) == 2


def test_get_dates_without_resampling(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "A"],
        [np.datetime64("2023-12-20", "ms"),  0,      10.0],
        [np.datetime64("2023-12-20", "ms"),  1,      12.0],
        [np.datetime64("2023-12-21", "ms"),  1,      13.0],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    all_realizations = provider.realizations()
    assert len(all_realizations) == 2

    all_dates = provider.dates(resampling_frequency=None)
    assert len(all_dates) == 1
    assert isinstance(all_dates[0], datetime)

    r0_dates = provider.dates(resampling_frequency=None, realizations=[0])
    r1_dates = provider.dates(resampling_frequency=None, realizations=[1])
    assert len(r0_dates) == 1
    assert len(r1_dates) == 2


def test_get_dates_with_daily_resampling(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "A",],
        [np.datetime64("2020-01-01", "ms"),  0,      10.0],
        [np.datetime64("2020-01-04", "ms"),  0,      40.0],
        [np.datetime64("2020-01-06", "ms"),  1,      60.0],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    all_realizations = provider.realizations()
    assert len(all_realizations) == 2

    all_dates = provider.dates(resampling_frequency=Frequency.DAILY)
    assert len(all_dates) == 6
    assert isinstance(all_dates[0], datetime)
    assert all_dates[0] == datetime(2020, 1, 1)
    assert all_dates[1] == datetime(2020, 1, 2)
    assert all_dates[4] == datetime(2020, 1, 5)
    assert all_dates[5] == datetime(2020, 1, 6)

    r0_dates = provider.dates(resampling_frequency=Frequency.DAILY, realizations=[0])
    assert len(r0_dates) == 4

    r1_dates = provider.dates(resampling_frequency=Frequency.DAILY, realizations=[1])
    assert len(r1_dates) == 1


def test_get_vector_metadata(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "A",  "B_r",  "C_t",   "D_r_t"],
        [np.datetime64("2023-12-20", "ms"),  0,      1.0,  10.0,   21.0,    31.0 ],
        [np.datetime64("2023-12-20", "ms"),  1,      2.0,  12.0,   22.0,    32.0 ],
        [np.datetime64("2023-12-21", "ms"),  1,      3.0,  13.0,   23.0,    33.0 ],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    meta = provider.vector_metadata("A")
    assert meta is None

    meta = provider.vector_metadata("B_r")
    assert meta and meta.is_rate is True

    meta = provider.vector_metadata("C_t")
    assert meta and meta.is_total is True

    meta = provider.vector_metadata("D_r_t")
    assert meta and meta.is_rate is True
    assert meta and meta.is_total is True


def test_get_vectors_without_resampling(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "A",  "B"],
        [np.datetime64("2023-12-20", "ms"),  0,      10.0,  21.0],
        [np.datetime64("2023-12-20", "ms"),  1,      12.0,  22.0],
        [np.datetime64("2023-12-21", "ms"),  1,      13.0,  23.0],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    all_vecnames = provider.vector_names()
    assert len(all_vecnames) == 2

    vecdf = provider.get_vectors_df(["A"], resampling_frequency=None)
    assert vecdf.shape == (3, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "A"]

    sampleddate = vecdf["DATE"][0]
    assert isinstance(sampleddate, datetime)

    vecdf = provider.get_vectors_df(["A"], resampling_frequency=None, realizations=[1])
    assert vecdf.shape == (2, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "A"]

    vecdf = provider.get_vectors_df(
        ["B", "A"], resampling_frequency=None, realizations=[0]
    )
    assert vecdf.shape == (1, 4)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "B", "A"]


def test_get_vectors_with_daily_resampling(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "TOT_t",  "RATE_r"],
        [np.datetime64("2020-01-01", "ms"),  1,      10.0,     1.0],
        [np.datetime64("2020-01-04", "ms"),  1,      40.0,     4.0],
        [np.datetime64("2020-01-06", "ms"),  1,      60.0,     6.0],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    vecdf = provider.get_vectors_df(
        ["TOT_t", "RATE_r"], resampling_frequency=Frequency.DAILY
    )

    date_arr = vecdf["DATE"].to_numpy()
    assert date_arr[0] == np.datetime64("2020-01-01", "ms")
    assert date_arr[1] == np.datetime64("2020-01-02", "ms")
    assert date_arr[2] == np.datetime64("2020-01-03", "ms")
    assert date_arr[3] == np.datetime64("2020-01-04", "ms")
    assert date_arr[4] == np.datetime64("2020-01-05", "ms")
    assert date_arr[5] == np.datetime64("2020-01-06", "ms")

    # Check interpolation for the total column
    tot_arr = vecdf["TOT_t"].to_numpy()
    assert tot_arr[0] == 10
    assert tot_arr[1] == 20
    assert tot_arr[2] == 30
    assert tot_arr[3] == 40
    assert tot_arr[4] == 50
    assert tot_arr[5] == 60

    # Check backfill for the rate column
    tot_arr = vecdf["RATE_r"].to_numpy()
    assert tot_arr[0] == 1
    assert tot_arr[1] == 4
    assert tot_arr[2] == 4
    assert tot_arr[3] == 4
    assert tot_arr[4] == 6
    assert tot_arr[5] == 6


def test_get_vectors_for_date_without_resampling(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "A",   "B",   "C"],
        [np.datetime64("2023-12-20", "ms"),  0,      10.0,  21.0,  31.0 ],
        [np.datetime64("2023-12-20", "ms"),  1,      12.0,  22.0,  32.0 ],
        [np.datetime64("2023-12-21", "ms"),  1,      13.0,  23.0,  33.0 ],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    all_dates = provider.dates(resampling_frequency=None)
    assert len(all_dates) == 1

    date_to_get = all_dates[0]
    assert isinstance(date_to_get, datetime)

    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A"])
    assert vecdf.shape == (2, 2)
    assert vecdf.columns.tolist() == ["REAL", "A"]

    date_to_get = all_dates[0]
    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A", "B"], [0])
    assert vecdf.shape == (1, 3)
    assert vecdf.columns.tolist() == ["REAL", "A", "B"]

    date_to_get = all_dates[0]
    vecdf = provider.get_vectors_for_date_df(date_to_get, ["A", "C"], [0])
    assert vecdf.shape == (1, 3)
    assert vecdf.columns.tolist() == ["REAL", "A", "C"]


def test_get_vectors_for_date_with_resampling(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                            "REAL",  "TOT_t",  "RATE_r"],
        [np.datetime64("2020-01-01", "ms"),  1,      10.0,     1.0],
        [np.datetime64("2020-01-04", "ms"),  1,      40.0,     4.0],
        [np.datetime64("2020-01-06", "ms"),  1,      60.0,     6.0],
    ]
    # fmt:on
    provider = _create_provider_obj_with_data(input_data, tmp_path)

    date_to_get = datetime(2020, 1, 3)

    df = provider.get_vectors_for_date_df(date_to_get, ["TOT_t", "RATE_r"])
    assert df.shape == (1, 3)

    assert df["REAL"][0] == 1
    assert df["TOT_t"][0] == 30.0
    assert df["RATE_r"][0] == 4.0


def test_monotonically_increasing_date_util_functions() -> None:
    table_with_duplicate = pa.Table.from_pydict(
        {
            "DATE": [
                np.datetime64("2020-01-01", "ms"),
                np.datetime64("2020-01-02", "ms"),
                np.datetime64("2020-01-02", "ms"),
                np.datetime64("2020-01-03", "ms"),
            ],
        },
    )

    table_with_decrease = pa.Table.from_pydict(
        {
            "DATE": [
                np.datetime64("2020-01-01", "ms"),
                np.datetime64("2020-01-05", "ms"),
                np.datetime64("2020-01-04", "ms"),
                np.datetime64("2020-01-10", "ms"),
            ],
        },
    )

    assert not _is_date_column_monotonically_increasing(table_with_duplicate)
    offending_pair = _find_first_non_increasing_date_pair(table_with_duplicate)
    assert offending_pair[0] == np.datetime64("2020-01-02", "ms")
    assert offending_pair[1] == np.datetime64("2020-01-02", "ms")

    assert not _is_date_column_monotonically_increasing(table_with_decrease)
    offending_pair = _find_first_non_increasing_date_pair(table_with_decrease)
    assert offending_pair[0] == np.datetime64("2020-01-05", "ms")
    assert offending_pair[1] == np.datetime64("2020-01-04", "ms")


def test_create_with_repeated_dates(tmp_path: Path) -> None:
    # fmt:off
    input_data = [
        ["DATE",                                 "REAL", "A"],
        [np.datetime64("2000-01-02T00:00", "ms"), 1,     10.0],
        [np.datetime64("2500-12-20T23:59", "ms"), 1,     11.0],
        [np.datetime64("2500-12-20T23:59", "ms"), 1,     12.0],
    ]
    # fmt:on

    with pytest.raises(ValueError):
        _create_provider_obj_with_data(input_data, tmp_path)



from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import xtgeo

from webviz_subsurface._providers import (
    EnsembleSurfaceProvider,
    QualifiedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
    SurfaceImageMeta,
    SurfaceImageServer,
)
from webviz_subsurface._providers.ensemble_surface_provider.ensemble_surface_provider import (
    SurfaceStatistic,
)
from webviz_subsurface.plugins._co2_leakage._utilities.plume_extent import (
    truncate_surfaces,
)


@dataclass
class TruncatedSurfaceAddress:
    name: str
    datestr: str
    realizations: List[int]
    basis_attribute: str
    threshold: float
    smoothing: float

    @property
    def attribute(self) -> str:
        return f"Truncated_{self.basis_attribute}_{self.threshold}_{self.smoothing}"


def publish_and_get_surface_metadata(
    server: SurfaceImageServer,
    provider: EnsembleSurfaceProvider,
    address: Union[SurfaceAddress, TruncatedSurfaceAddress],
) -> Tuple[Optional[SurfaceImageMeta], str]:
    if isinstance(address, TruncatedSurfaceAddress):
        return _publish_and_get_truncated_surface_metadata(server, provider, address)
    provider_id: str = provider.provider_id()
    qualified_address = QualifiedSurfaceAddress(provider_id, address)
    surf_meta = server.get_surface_metadata(qualified_address)
    if not surf_meta:
        # This means we need to compute the surface
        surface = provider.get_surface(address)
        if not surface:
            raise ValueError(f"Could not get surface for address: {address}")
        server.publish_surface(qualified_address, surface)
        surf_meta = server.get_surface_metadata(qualified_address)
    return surf_meta, server.encode_partial_url(qualified_address)


def _publish_and_get_truncated_surface_metadata(
    server: SurfaceImageServer,
    provider: EnsembleSurfaceProvider,
    address: TruncatedSurfaceAddress,
) -> Tuple[Optional[SurfaceImageMeta], str]:
    qualified_address = QualifiedSurfaceAddress(
        provider.provider_id(),
        # TODO: Should probably use a dedicated address type for this. Statistical surface
        #  is the closest, as it allows including a list of realizations. However, it is
        #  perhaps not very "statistical", and the provided SurfaceStatistic is not
        #  appropriate here.
        StatisticalSurfaceAddress(
            address.attribute,
            address.name,
            address.datestr,
            SurfaceStatistic.MEAN,
            address.realizations,
        ),
    )
    surf_meta = server.get_surface_metadata(qualified_address)
    if surf_meta is None:
        surface = _generate_surface(provider, address)
        if surface is None:
            raise ValueError(f"Could not generate surface for address: {address}")
        server.publish_surface(qualified_address, surface)
        surf_meta = server.get_surface_metadata(qualified_address)
    return surf_meta, server.encode_partial_url(qualified_address)


def _generate_surface(
    provider: EnsembleSurfaceProvider,
    address: TruncatedSurfaceAddress,
) -> Optional[xtgeo.RegularSurface]:
    surfaces = [
        provider.get_surface(
            SimulatedSurfaceAddress(
                attribute=address.basis_attribute,
                name=address.name,
                datestr=address.datestr,
                realization=r,
            )
        )
        for r in address.realizations
    ]
    surfaces = [s for s in surfaces if s is not None]
    if len(surfaces) == 0:
        return None
    plume_count = truncate_surfaces(surfaces, address.threshold, address.smoothing)
    template: xtgeo.RegularSurface = surfaces[0].copy()  # type: ignore
    template.values = plume_count
    template.values.mask = plume_count < 1e-4  # type: ignore
    return template



import numpy as np
import xtgeo
from webviz_config.common_cache import CACHE


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_cube_data(cube_path: str) -> xtgeo.Cube:
    return xtgeo.cube_from_file(cube_path)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_xline(cube: xtgeo.Cube, xline: int) -> np.ndarray:
    idx = np.where(cube.xlines == xline)
    return cube.values[:, idx, :][:, 0, 0].T


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_iline(cube: xtgeo.Cube, iline: int) -> np.ndarray:
    idx = np.where(cube.ilines == iline)
    return cube.values[idx, :, :][0, 0, :].T


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_zslice(cube: xtgeo.Cube, zslice: float) -> np.ndarray:
    idx = np.where(cube.zslices == zslice)
    return cube.values[:, :, idx][:, :, 0, 0].T
